/*

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Copyright (C) 2006 by Clemens Helfmeier,
    Clemens(at)Helfmeier[dot]de

*/
/**
 * Startup Code for ARM
 *
 * Assembler file for DAB
 * this file handles the very basic aspects of
 * the DAB-Interrupt. It will call an external c-function for
 * handling the DAB-event on a higher level (Debugging, trace, etc.)
 *
 * by: Clemens Helfmeier
 * ClemensHelfmeier {at] gmx `dot" de
 * (replace the corresponding at and dot in the email, please).
 */
#define ASSEMBLER

#include "AT91SAM7S64.inc"
#include "arm.inc"
#include "startup_config.h"

.text
.func DAB_Handler
.global DAB_Handler
DAB_Handler:
    /* this handles the data Abort interrupt
     * R13 contains the dab stack pointer,
     * R14 contains the return address
     *
     * stack needed in DAB mode: r0,r1,lr,spsr = 16 bytes
     */

    /* save working register */
    stmfd   sp!, {r0-r2}

    
    /* copy the lr to r0 and take it over into the System mode (the registers lr is banked.) */
    sub r0, lr, #8
    
    /* disable Interrupts (no nested DABs, no irq's while DABs), switch to system mode */
    mrs r2, CPSR        /* we must read the whole CPSR, but we can write back only the low part */
    and r2, r2, #(ARM_MODE_System)
    orr r2, r2, #(ARM_MODE_System | ARM_F_BIT | ARM_I_BIT)
    msr CPSR_c, r2

    /* save registers in user mode or System mode stack (which is the same)*/
    stmfd   sp!, {r3-r12,r14}
    
    /* call service routine with arguments */
    /* R0 = source address (abort mode lr) */
    /* R1 = stack pointer (system mode stack) */
    /* R2 = real frame pointer*/
    mov r1, sp
    mov r2, r9
    CALL_FUNC data_abort_handler
    /* the new return address is in r0 */

    /* restore registers that where saved in user/system mode stack */
    ldmia   sp!, {r3-r12,r14}
    
    /* switch back to Abort mode */
    mrs r2, CPSR        /* we must read the whole CPSR, but we can write back only the low part */
    and r2, r2, #(ARM_MODE_Abort)
    orr r2, r2, #(ARM_MODE_Abort)
    msr CPSR_c, r2
    
    /* copy the new retun address from r0 to lr */
    mov lr, r0
            
    /* restore environment r0,r1,r2 from abort mode stack */
    ldmia   sp!, {r0-r2}

    /* return from DAB-Handler, adjust the value of return address, copy spsr to cpsr */
    subs pc, lr, #0

.equ    DAB_Handler_Size, ( . - DAB_Handler )

.global DAB_Handler_Size
.size DAB_Handler, . - DAB_Handler
.endfunc


.text
.func PAB_Handler
.global PAB_Handler
PAB_Handler:
    /* this handles the prefetch Abort interrupt
     * R13 contains the pab stack pointer,
     * R14 contains the return address
     *
     * stack needed in PAB mode: r0,r1,lr,spsr = 16 bytes
     */

    /* save working register */
    stmfd   sp!, {r0-r1}

    /* copy lr from prefetch abort mode. */
    sub r0, lr, #4

    /* disable Interrupts (no nested DABs, no irq's while PABs), switch to system mode */
    mrs r1, CPSR        /* we must read the whole CPSR, but we can write back only the low part */
    and r1, r1, #(ARM_MODE_System)
    orr r1, r1, #(ARM_MODE_System | ARM_F_BIT | ARM_I_BIT)
    msr CPSR_c, r1

    /* save registers in user mode or System mode stack (which is the same)*/
    stmfd   sp!, {r2-r12,r14}
    
    /* call service routine with arguments */
    /* R0 = source address (prefetch abort mode)*/
    /* R1 = stack pointer (system mode) */
    /* R2 = real frame pointer*/
    mov r2, r9
    mov r1, sp
    CALL_FUNC prefetch_abort_handler
    /* the new return address is in r0 */

    /* restore registers that where saved in user/system mode stack */
    ldmia   sp!, {r2-r12,r14}
    
    /* copy new lr from r0 */
    mov lr, r0
            
    /* restore environment r0,r1,lr,spsr*/
    ldmia   sp!, {r0-r1}

    /* return from DAB-Handler, adjust the value of return address, copy spsr to cpsr */
    subs pc, lr, #0

.equ    PAB_Handler_Size, ( . - PAB_Handler )

.global PAB_Handler_Size
.size PAB_Handler, . - PAB_Handler
.endfunc


