/*

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Copyright (C) 2006 by Clemens Helfmeier,
    Clemens(at)Helfmeier[dot]de

*/

/** @file spi.h
 * @brief SPI-Library
 * @author Clemens Helfmeier
 * @date 10.11.2006
 * 
 * 
 * SPI Library.
 * There are two types of SPI functions: synchronous and asynchronous functions. Synchronous functions
 * will always return after the transfer is finished, asynchronous functions will return immediatly after
 * starting or initiating a transfer. Always call SPI_SYNCHRONIZE before using the synchronous functions.
 * In the SPI interrupt SPI_SYNCHRONIZE doesn't need to be called, the interrupt is already synchronous.
 * There is no point in doing short transfers via interrupt control, here it would be better to use the
 * synchronous functions and block the processor for that time. Interrupt handling needs much time since
 * the whole register set needs to be pushed on the stack.
 * 
 */


#ifndef _SPI_H_
#define _SPI_H_

#include "AT91SAM7S64.h"
#define __inline extern inline
#include "lib_AT91SAM7S64.h"

#include "system/error.h"
#include "controller/config_spi.h"
#include "controller/compiler.h"

#ifdef __cplusplus
extern "C" {
#endif

    /** the spi interrupt should be defined in the module that handles
     * an empty spi bus.
     */
void spi_interrupt_handler( void );
    
    
extern const unsigned int cs_modes[ 4 ];

    
static inline unsigned int SPI_SYNCHRONIZE( void ) {
    unsigned int ret = (AT91C_BASE_AIC->AIC_IMR & (1<<AT91C_ID_SPI));
    AT91C_BASE_AIC->AIC_IDCR = (1<<AT91C_ID_SPI);
    return ret;
}
static inline void SPI_DESYNCHRONIZE( unsigned int en ) {
    if (en == (1<<AT91C_ID_SPI)) {
         AT91C_BASE_AIC->AIC_IECR = (1<<AT91C_ID_SPI);
    }
}

// these functions should be used when there is no transfer to be handled
// or when new transfers are possible.
// Here, we _MUST_ use TXEMPTY interrupt because this is the only interrupt that
// is active if the SPI interface is really idle! Any other interrupt would not be
// synchronous. Now we can use the spi_sync_* functions.
static inline void SPI_ENABLE_INTERRUPT( void ) {
    AT91C_BASE_SPI->SPI_IER = AT91C_SPI_TXEMPTY;
}
static inline void SPI_DISABLE_INTERRUPT( void ) {
    AT91C_BASE_SPI->SPI_IDR = AT91C_SPI_TXEMPTY;
}

/** initialize the spi interface */
ERROR spi_init( void );

/** synchronous functions. they use polling transfers */
static inline void spi_sync_tmit( unsigned short data ) {
    while ((AT91C_BASE_SPI->SPI_SR & AT91C_SPI_TDRE) == 0);
    AT91C_BASE_SPI->SPI_TDR = data & 0xFFFF;
}

static inline unsigned short spi_sync_rmit( void ) {
    while ((AT91C_BASE_SPI->SPI_SR & AT91C_SPI_RDRF) == 0);
    return AT91C_BASE_SPI->SPI_RDR & 0xFFFF;
}

void spi_sync_start( unsigned int slave );

static inline void spi_sync_end( void ) {
    // wait for any transfer to finish
    while ((AT91C_BASE_SPI->SPI_SR & AT91C_SPI_TXEMPTY) == 0);
    AT91C_BASE_SPI->SPI_MR = (AT91C_BASE_SPI->SPI_MR & (~AT91C_SPI_PCS)) | ((~((0)<<16)) & AT91C_SPI_PCS);
    // release cs lines
    AT91F_PIO_SetOutput( AT91C_BASE_PIOA, SPI_CS_0 | SPI_CS_1 | SPI_CS_2 | SPI_CS_3 );
}

/** asynchronous functions, they do not block the processor. DMA transfers */
static inline unsigned int spi_async_idle( void ) {
    return (AT91C_BASE_SPI->SPI_SR & AT91C_SPI_TXEMPTY);
}
static inline void spi_async_xfer( void * txdata, void * rxdata, unsigned int len ) {
    // disable transfers:
    AT91C_BASE_PDC_SPI->PDC_PTCR = (AT91C_PDC_TXTDIS | AT91C_PDC_RXTDIS);
    AT91C_BASE_PDC_SPI->PDC_TCR = len;      // tx is always needed
    AT91C_BASE_PDC_SPI->PDC_TPR = (int)txdata;   // if this is 0, there will be transmitted random data, no need to worry.
    if (rxdata != 0) {
        AT91C_BASE_PDC_SPI->PDC_RCR = len;
        AT91C_BASE_PDC_SPI->PDC_RPR = (int)rxdata;
        AT91C_BASE_PDC_SPI->PDC_PTCR = (AT91C_PDC_TXTEN | AT91C_PDC_RXTEN);
    } else {
        AT91C_BASE_PDC_SPI->PDC_RCR = 0;
        AT91C_BASE_PDC_SPI->PDC_PTCR = (AT91C_PDC_TXTEN);
    }
    // enable (start) transfers
}

#ifdef __cplusplus
}
#endif

#endif
