/*

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Copyright (C) 2006 by Clemens Helfmeier,
    Clemens(at)Helfmeier[dot]de

*/

/** @file debug.h
 * @author Clemens Helfmeier
 * @date 25.06.06
 * @brief This file contains debuging defines and prototypes for the functions used by the defines.
 *
 * Debuging is achieved by sending messages over the USART to a terminal. Debuging code is only compiled,
 * if the message level is higher or equal to the \c DBG_LEVEL. This file must be included in a sepcial
 * way:
 * \code
 * #include "debug.h"
 * #define DBG_LEVEL	DBG_LEVEL_ERROR
 * #include "debug.h"
 * \endcode
 *
 * The first include will define the \c DBG_LEVEL_* values, the second one the \c DBG_*_*() defines.
 *
 * Since the documentation of this file is a bit difficult, the useful functions are listed here:
 * 
 * DBG_C_S, DBG_C_B, DBG_C_I: Send a critical message to the debuging output (string, byte, integer) \n
 * DBG_E_S, DBG_E_B, DBG_E_I: Send an error message to the debuging output (string, byte, integer) \n
 * DBG_I_S, DBG_I_B, DBG_I_I: Send an information message to the debuging output (string, byte, integer) \n
 * DBG_V_S, DBG_V_B, DBG_V_I: Send a verbose message to the debuging output (string, byte, integer) \n
 */

/* on the first include, define all the DBG_LEVEL_* defines. If they are already defined, they won't be
 * redefined because it is the second include.
 */
/// @brief Only show critical events
#ifndef DBG_LEVEL_CRITICAL
#define DBG_LEVEL_CRITICAL	10
#endif

/// @brief Show critical events and errors that are not fatal.
#ifndef DBG_LEVEL_ERROR
#define DBG_LEVEL_ERROR		8
#endif

/// @brief Show information, too.
#ifndef DBG_LEVEL_INFO
#define DBG_LEVEL_INFO		4
#endif

/// @brief Show really everything.
#ifndef DBG_LEVEL_VERBOSE
#define DBG_LEVEL_VERBOSE		2
#endif

/* if the DBG_LEVEL define is defined, it is the second include, so the DBG_*_* defines should be placed...
 */
#ifdef DBG_LEVEL

/* this part should only be included once. */
#ifndef _DEBUG_H_
#define _DEBUG_H_

#ifdef __cplusplus
extern "C" {
#endif

#define PSTR( str )    str
// (__extension__({static char __c[] __attribute__ ((section (".text"))) = (str); &__c[0];}))

#if DBG_LEVEL <= DBG_LEVEL_CRITICAL
#define DBG_C_S( string )		dbg_puts( PSTR(string) )
#define DBG_C_B( integer )		dbg_putb( integer )
#define DBG_C_I( integer )		dbg_puti( integer )
#define DBG_C_P( pointer )      { int i; memcpy( &i, &pointer, 4); dbg_puti( i ); }
#else
#define DBG_C_S( string )
#define DBG_C_B( integer )
#define DBG_C_I( integer )
#define DBG_C_P( pointer )
#endif

#if DBG_LEVEL <= DBG_LEVEL_ERROR
#define DBG_E_S( string )		dbg_puts( PSTR(string) )
#define DBG_E_B( integer )		dbg_putb( integer )
#define DBG_E_I( integer )		dbg_puti( integer )
#define DBG_E_P( pointer )      { int i; memcpy( &i, &pointer, 4); dbg_puti( i ); }
#else
#define DBG_E_S( string )
#define DBG_E_B( integer )
#define DBG_E_I( integer )
#define DBG_E_P( pointer )
#endif

#if DBG_LEVEL <= DBG_LEVEL_INFO
#define DBG_I_S( string )		dbg_puts( PSTR(string) )
#define DBG_I_B( integer )		dbg_putb( integer )
#define DBG_I_I( integer )		dbg_puti( integer )
#define DBG_I_P( pointer )      { int i; memcpy( &i, &pointer, 4); dbg_puti( i ); }
#else
#define DBG_I_S( string )
#define DBG_I_B( integer )
#define DBG_I_I( integer )
#define DBG_I_P( pointer )
#endif

#if DBG_LEVEL <= DBG_LEVEL_VERBOSE
#define DBG_V_S( string )		dbg_puts( PSTR(string) )
#define DBG_V_B( integer )		dbg_putb( integer )
#define DBG_V_I( integer )		dbg_puti( integer )
#define DBG_V_P( pointer )      { int i; memcpy( &i, &pointer, 4); dbg_puti( i ); }
#else
#define DBG_V_S( string )
#define DBG_V_B( integer )
#define DBG_V_I( integer )
#define DBG_V_P( pointer )
#endif



void dbg_puts( char* s );
void dbg_putb( char x );
void dbg_puti( int x );
void dbg_init( void );

#ifdef __cplusplus
}
#endif

// _DBG_H_
#endif

// DBG_LEVEL
#endif
