/*

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Copyright (C) 2006 by Clemens Helfmeier,
    Clemens(at)Helfmeier[dot]de

*/


#include "mem.h"

#include <string.h>
#include "utils/synchronization.h"


// debugging support
#include "debug/debug.h"
#define DBG_LEVEL   DBG_LEVEL_ERROR
#include "debug/debug.h"


/** memory implementation
 * the memory is totally unorganized. the only structure that contains all information
 * about the memory is the mem_buffer. Each memory block - no matter if it is free 
 * or already allocated - is lead by exactly one of this structure, this gives an overhead
 * of 12 bytes per allocated memory block.
 * free buffers are kept in a chain, starting at first_free. The structure can still be optimized
 * if for free buffers, the "used" field - which is then not used - is used for the "next" field (union)
 */

#define MEM_BYTE_COUNT              (50000)

char mem_mem[ MEM_BYTE_COUNT ];

typedef union {
    struct {
        unsigned int len;
        unsigned int used;
    } allocated;
    struct {
        unsigned int len;
        void * next;
    } free;
} mem_buffer;


mem_buffer * first_free;

/** synchronization object for any functions accessing the memory
 * buffer structures.
 */
SYNC_OBJECT( MEM_SYNC );


#if 1
#include "AT91SAM7S64.h"
#define __inline extern inline
#include "lib_AT91SAM7S64.h"
#define BENCHMARK_START()           AT91F_PIO_SetOutput( AT91C_BASE_PIOA, (1<<10) )
#define BENCHMARK_STOP()            AT91F_PIO_ClearOutput( AT91C_BASE_PIOA, (1<<10) )
#define BENCHMARK_INIT()            AT91F_PIO_CfgOutput( AT91C_BASE_PIOA, (1<<10) )
#else
#define BENCHMARK_START()
#define BENCHMARK_STOP()
#define BENCHMARK_INIT()
#endif


void mem_init( void ) {
    SYNC( MEM_SYNC );    
    memset( mem_mem, 0x00, sizeof(mem_mem) );
    BENCHMARK_INIT();
    // all memory is free:
    first_free = (mem_buffer*)mem_mem;
    first_free->free.len = sizeof(mem_mem);
    first_free->free.next = 0;
    DESYNC( MEM_SYNC );
}

void mem_use( void * mem ) {
    SYNC( MEM_SYNC );
    BENCHMARK_START();
    mem_buffer * mb = (mem_buffer*)((unsigned int)mem - sizeof(mem_buffer));
    // check if mb is in the free queue, if so give an error:
    mem_buffer * fm = first_free;
    while (fm != 0) {
        if (mb == fm) {
            DBG_C_S( "mem_use: trying to use not allocated memory.\n" );
            while(1);
        }
        fm = fm->free.next;
    }
    if (((unsigned int)mb < (unsigned int)mem_mem) || (((unsigned int)mb) + mb->allocated.len > (unsigned int)mem_mem + sizeof(mem_mem))) {
        DBG_C_S( "mem_use: invalid memory handle given: " ); DBG_C_I( mb ); DBG_C_S( ".\n" );
        while(1);
    }
    DBG_I_S( "mem_use: " ); DBG_I_I( mb ); DBG_I_S( ".\n" );
    mb->allocated.used++;
    BENCHMARK_STOP();
    DESYNC( MEM_SYNC );
}
void mem_free( void * mem ) {
    SYNC( MEM_SYNC );
    BENCHMARK_START();
    mem_buffer * mb = (mem_buffer*)((unsigned int)mem - sizeof(mem_buffer));
    // check if memory is allocated
    mem_buffer * fm = first_free;
    while (fm != 0) {
        if (mb == fm) {
            DBG_C_S( "mem_free: trying to free not allocated memory.\n" );
            while(1);
        }
        fm = fm->free.next;
    }
    if (((unsigned int)mb < (unsigned int)mem_mem) || (((unsigned int)mb) + mb->allocated.len > (unsigned int)mem_mem + sizeof(mem_mem))) {
        DBG_C_S( "mem_free: invalid memory handle given: " ); DBG_C_I( mb ); DBG_C_S( ".\n" );
        while(1);
    }
    DBG_I_S( "mem_free: " ); DBG_I_I( mb ); DBG_I_S( ".\n" );
    mb->allocated.used--;
    if (mb->allocated.used <= 0) {
        /* really free the buffer */
        // find a free memory buffer that touches the one, we are freeing
        mem_buffer * fm = first_free, * lm = 0;
        mem_buffer * pm = 0, * nm = 0;  // previous and next memory
        mem_buffer * ppm, * nnm;
        while (fm != 0) {
            if ((mem_buffer*)((unsigned int)fm + fm->free.len) == mb) {
                if (pm != 0) {
                    DBG_E_S( "mem_free: should concatenate two previous buffers!\n" );
                } else {
                    DBG_V_S( "mem_free: concatenating previous buffer at " ); DBG_V_I( fm ); DBG_V_S( ".\n" );
                    pm = fm;
                    ppm = lm;
                }
            }
            if (fm == (mem_buffer*)((unsigned int)mb + mb->free.len)) {
                if (nm != 0) {
                    DBG_E_S( "mem_free: should concatenate two next buffers!\n" );
                } else {
                    DBG_V_S( "mem_free: concatenating next buffer at " ); DBG_V_I( fm ); DBG_V_S( ".\n" );
                    nm = fm;
                    nnm = lm;
                }
            }
            lm = fm;
            fm = fm->free.next;
        }
        // now check for the memory buffer that is in the new free-queue:
        if ((pm == 0) && (nm == 0)) {
            // no concatenation, simply add mb.
            DBG_V_S( "mem_free: no concatenation at " ); DBG_V_I( mb ); DBG_V_S( ", " ); DBG_V_I( mb->free.len ); DBG_V_S( ".\n" );
            mb->free.next = first_free;
            first_free = mb;
        } else if (pm == 0) {
            // only nm has been concatenated. let nnm->next point to mb where the new buffer is and also copy
            // the nm->next pointer to mb->next
            mb->free.len += nm->free.len;
            mb->free.next = nm->free.next;
            if (nnm == 0) {
                // nm is the first entry in the queue:
                first_free = mb;
            } else {
                // nm is some following entry:
                nnm->free.next = mb;
            }
        } else if (nm == 0) {
            // only pm has been concatenated. this is quite easy, there is nothing to be done since
            // pm is already in the queue and just has been increased in size.
            pm->free.len += mb->free.len;
        } else {
            // nm and pm has been concatenated. now we need to resize pm to contain nm and pm and mb
            pm->free.len += mb->free.len + nm->free.len;
            // nm is not existant any more, we need to modify nnm->next
            if (nnm == 0) {
                // nm was the first entry in the list
                if (first_free != nm) {
                    DBG_C_S( "mem_free: invalid nnm value: " ); DBG_C_I( nnm ); DBG_C_S( ".\n" );
                    while(1);
                }
                first_free = nm->free.next;
            } else {
                nnm->free.next = nm->free.next;
            }
        }
    } else {
        DBG_V_S( "mem_free: not freeing buffer at " ); DBG_V_I( mb ); DBG_V_S( ", " ); DBG_V_I( mb->free.len ); DBG_V_S( ".\n" );
    }
    BENCHMARK_STOP();
    DESYNC( MEM_SYNC );
}
unsigned int mem_usecount( void * mem ) {
    mem_buffer * mb = (mem_buffer*)((unsigned int)mem - sizeof(mem_buffer));
    return mb->allocated.used;
}

void * mem_allocate( unsigned int size ) {
    SYNC( MEM_SYNC );
    BENCHMARK_START();
    // allocate a new memory buffer
    // allocate the mem_buffer struct, too
    // align size by 4 bytes
    size = (size + sizeof( mem_buffer ) + 3) & 0xFFFFFFFC;
    mem_buffer * mb = first_free, *ob = 0, *nb;
    while ((mb != 0) && (mb->free.len < size)) {
        ob = mb;
        mb = mb->free.next;
    }
    if (mb == 0) {
        // allocation failed!
        DBG_E_S( "mem_allocate: out of memory.\n" );
        mb = (mem_buffer*)(-sizeof(mem_buffer));
    } else {
        DBG_V_S( "mem_allocate: found memory at " ); DBG_V_I( mb ); DBG_V_S( " with len " ); DBG_V_I( mb->free.len ); DBG_V_S( ".\n" );
        // calculate the rest of the free buffer space:
        if (mb->free.len > size + sizeof( mem_buffer )) {
            // only do a new buffer if that one is big enough to contain at least the mem_buffer structure.
            nb = (mem_buffer *)((unsigned int)mb + size);
            nb->free.next = mb->free.next;
            nb->free.len = mb->free.len - size;
            DBG_V_S( "mem_allocate: splitting memory: (" ); DBG_V_I( mb ); DBG_V_S( ", " ); DBG_V_I( size ); DBG_V_S( "), (" ); DBG_V_I( nb ); DBG_V_S( ", " ); DBG_V_I( nb->free.len ); DBG_V_S( ").\n" );
        } else {
            DBG_V_S( "mem_allocate: no splitting.\n" );
            nb = mb->free.next;
        }
        if (ob != 0) {
            // add the new free buffer to the free buffer chain:
            DBG_V_S( "mem_allocate: new buffer is in the middle of the chain, previous: " ); DBG_V_I( ob ); DBG_V_S( ".\n" );
            ob->free.next = nb;
        } else {
            // the new buffer index is the first one:
            DBG_V_S( "mem_allocate: new buffer is the first buffer in the chain.\n" );
            first_free = nb;
        }
        // create the new used buffer
        mb->allocated.len = size;
        mb->allocated.used = 1;
    }
    DBG_I_S( "mem_allocate: " ); DBG_I_I( mb ); DBG_I_S( ".\n" );
    BENCHMARK_STOP();
    DESYNC( MEM_SYNC );
    return (void*)((unsigned int)mb + sizeof(mem_buffer));
}

unsigned int mem_fill( void ) {
    SYNC( MEM_SYNC );
    BENCHMARK_START();
    // return amount of used memory
    unsigned int len = 0, i = 0;
    mem_buffer * mb = first_free;
    while (mb != 0) {
        len += mb->free.len;
        mb = mb->free.next;
        i++;
    }
    DBG_I_S( "mem_fill: number of free buffers: " ); DBG_I_I( i ); DBG_I_S( ".\n" );
    BENCHMARK_STOP();
    DESYNC( MEM_SYNC );
    return len;
}

void mem_debug( void ) {
    SYNC( MEM_SYNC );
    // print some debugging information about the memory buffers...
    mem_buffer * mb = first_free;
    while (mb != 0) {
        DBG_C_S( "mem_debug: free buffer at " ); DBG_C_I( mb ); DBG_C_S( ", " ); DBG_C_I( mb->free.len ); DBG_C_S( ".\n" );
        mb = mb->free.next;
    }
    mb = (mem_buffer*)mem_mem;
    while ((unsigned int)mb < (unsigned int)mem_mem + sizeof(mem_mem)) {
        DBG_C_S( "mem_debug: buffer at " ); DBG_C_I( mb ); DBG_C_S( ", " ); DBG_C_I( mb->free.len ); DBG_C_S( " -> " ); DBG_C_I( mb->free.next ); DBG_C_S( ".\n" );
        if (mb->free.len > sizeof(mem_mem)) {
            DBG_C_S( "mem_debug: invalid length field!" );
//            while(1);
        }
        if (mb->free.len == 0) {
            DBG_C_S( "mem_debug: corrupted packet length!" );
            mb++;
        } else {
            mb = (mem_buffer*)(((unsigned int)mb + mb->free.len + 3) & 0xFFFFFFFC);
        }
    }
    DESYNC( MEM_SYNC );
}
