/*

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Copyright (C) 2006 by Clemens Helfmeier,
    Clemens(at)Helfmeier[dot]de

*/




#include "enc28j60.h"
#include "net/netconfig.h"

#include "memory/mem.h"
#include "controller/compiler.h"

#include "enc28j60_register.h"
#include "system/system.h"

#include "debug/debug.h"
#define DBG_LEVEL   DBG_LEVEL_ERROR
#include "debug/debug.h"


/** Implementation of the ENC28j60.
 * All packets are processed asynchronously.
 * There are two fifo's, one for reception, one for transmission.
 */


typedef struct {
        void * mem;
        unsigned int len;
} ENC28j60_FIFO_ENTRY;
#define ENC28j60_FIFO_LEN       4
typedef struct {
    volatile unsigned int rpointer;
    volatile unsigned int len;
    ENC28j60_FIFO_ENTRY data[ ENC28j60_FIFO_LEN ];
} ENC28j60_FIFO;


/// enc28j60 tx packet fifo
ENC28j60_FIFO enc28j60_txfifo;
/// enc28j60 rx packet fifo
ENC28j60_FIFO enc28j60_rxfifo;

/** controller specific defines */
#include "AT91SAM7S64.h"
#define __inline extern inline
#include "lib_AT91SAM7S64.h"
void enc28j60_irq_handler( void );
#define ENC28j60_IRQ_SETUP()     {  \
    AT91F_AIC_CfgPMC();             \
    AT91F_AIC_ConfigureIt( AT91C_BASE_AIC, AT91C_ID_IRQ1, 3 /*priority*/, 0, enc28j60_irq_handler );         \
    AT91F_PIO_CfgPeriph( AT91C_BASE_PIOA, AT91C_PA30_IRQ1, 0 ); \
}
#define ENC28j60_IRQ_ENABLE()    { \
    AT91F_AIC_EnableIt( AT91C_BASE_AIC, AT91C_ID_IRQ1 );        \
}
#define ENC28j60_IRQ_DISABLE()   { \
    AT91F_AIC_DisableIt( AT91C_BASE_AIC, AT91C_ID_IRQ1 );        \
}
#if 0
#define BENCHMARK_START()           AT91F_PIO_SetOutput( AT91C_BASE_PIOA, (1<<10) )
#define BENCHMARK_STOP()            AT91F_PIO_ClearOutput( AT91C_BASE_PIOA, (1<<10) )
#define BENCHMARK_INIT()            AT91F_PIO_CfgOutput( AT91C_BASE_PIOA, (1<<10) )
#else
#define BENCHMARK_START()
#define BENCHMARK_STOP()
#define BENCHMARK_INIT()
#endif
         
/** ENC28j60 specific defines and variables */
#define RX_START    0x0000
#define RX_END      0x19FF
#define TX_START    0x1A00
#define TX_END      0x1FFF
#define MTU         1518
unsigned short enc28j60_rx_packet_pointer = 0;

unsigned int enc28j60_protect = 0;
#define ENC28j60_PROTECT_CHECK()        {   \
    if (arm_swap( 1, &enc28j60_protect) != 0) {            \
        DBG_C_S( "enc28j60 protection error (1).\n" ); \
    }                                       \
}

#define ENC28j60_PROTECT_UNCHECK()      {   \
    if (arm_swap( 0, &enc28j60_protect) != 1) {            \
        DBG_C_S( "enc28j60 protection error (0).\n" ); \
    }                                       \
}

unsigned int enc28j60_txfifo_fill( void ) {
    return enc28j60_txfifo.len;
}
unsigned int enc28j60_rxfifo_fill( void ) {
    return enc28j60_rxfifo.len;
}

void enc28j60_receive_packets( void ) {
    // test if the rx fifo has packets
    // no synchronization needed for reading the len since this is the only place where packets
    // can be removed from the rx_fifo!
    while (enc28j60_rxfifo.len != 0) {
        DBG_I_S( "enc28j60_receive_packets: packet in queue, handling ... " );
        // get the packet
        ENC28j60_FIFO_ENTRY p;
        // now synchronize with SPI (no packets can be recieved now.)
        unsigned int sync = SPI_SYNCHRONIZE();
        ENC28j60_PROTECT_CHECK();
        // get the packet from the fifo
        p = enc28j60_rxfifo.data[ enc28j60_rxfifo.rpointer++ ];
        if (enc28j60_rxfifo.rpointer >= ENC28j60_FIFO_LEN) enc28j60_rxfifo.rpointer = 0;
        enc28j60_rxfifo.len--;
        // ok, now the fifo is finished, desynchronize
        ENC28j60_PROTECT_UNCHECK();
        SPI_DESYNCHRONIZE( sync );
        // handle the packet now:
        MAC_INPUT_HANDLER( p.mem, p.len );
        // free the packet, we don't need it any more
        mem_free( p.mem );
        DBG_I_S( " done.\n" );
    }
    // now check for new packets
    //SPI_ENABLE_INTERRUPT();
    // this is automatically done by using the hardware interrupt line
}

/** Currently this may not be called from an interrupt. The spi handler won't
 * take a packet from the tx fifo if the fifo is ful -> endless waiting for tx fifo
 * becoming empty.
 */
ERROR enc28j60_send_packet( void * mem, unsigned int len ) {
    // send the packet, put it in the fifo
    // wait for a fifo place:
    unsigned int sync = 0;
    while (1) {
        sync = SPI_SYNCHRONIZE();
        ENC28j60_PROTECT_CHECK();
        if (enc28j60_txfifo.len < ENC28j60_FIFO_LEN) break;
        ENC28j60_PROTECT_UNCHECK();
        SPI_DESYNCHRONIZE( sync );
        // at least execute some instructions here!
        DBG_I_S( "enc28j60_send_packet: waiting for empty fifo ...\n" );
        // we must continously enable the interrupt here because if the enc is busy and the pdc is idel,
        // it will be disabled. but there is no way to determine if the interrupt may be enabled again when
        // the transmitting packet is out of the enc.
        while (enc28j60_txfifo.len >= ENC28j60_FIFO_LEN) SPI_ENABLE_INTERRUPT();

    }
    DBG_V_S( "enc28j60_send_packet: putting packet to txqueue.\n" );
    // ok, put the packet to the fifo now:
    unsigned int i = enc28j60_txfifo.rpointer + (enc28j60_txfifo.len++);
    if (i >= ENC28j60_FIFO_LEN) i -= ENC28j60_FIFO_LEN;
    enc28j60_txfifo.data[i].mem = mem;
    enc28j60_txfifo.data[i].len = len;
    // we use the packet now:
    mem_use( mem );
    // ok, enable packet transmission by enabling the spi interrupt
    SPI_ENABLE_INTERRUPT();
    // now desynchronize again:
    ENC28j60_PROTECT_UNCHECK();
    SPI_DESYNCHRONIZE( sync );
    // ok, successfull:
    return ERROR_SUCCESS;
}

ERROR enc28j60_init( void ) {
    // initialisation is taking place in synchronous spi mode
    // switch to synchronous mode:
    unsigned int sync = SPI_SYNCHRONIZE();
    ENC28j60_PROTECT_CHECK();
    // initialize the enc28j60. always use the spi_sync_* functions here!
    // clear packet buffers
    enc28j60_txfifo.rpointer = 0;
    enc28j60_txfifo.len = 0;
    enc28j60_rxfifo.rpointer = 0;
    enc28j60_rxfifo.len = 0;
    
    /* reset the controller */
    soft_reset();
    /* reset recieve and transmit logic */
    set_bits( ECON1, TXRST | RXRST );
    delay_us(100);
    clear_bits( ECON1, TXRST | RXRST );
    /* clear Mac Reset */
    write_mac_reg( MACON2, 0x00 );

    delay_ms( 20 );
    /* wait until clock is ready */
    unsigned char i;
    do {
        i = read_eth_reg( ESTAT );
    } while ((i & 0x08) || (~i & CLKRDY));
    /* wait for phy reset to finish */
    while ((read_phy_reg16(PHCON1) & PRST) != 0);
    /* we could do the memory test here now */
    /* initialize ECON2 (Clear power save) */
    write_eth_reg( ECON2, AUTOINC );
    DBG_I_S( "enc28j60_init(): EREVID: " ); DBG_I_B( read_eth_reg( EREVID ) ); DBG_I_S( "\n" );
    /* initialize the recieve buffer, this should be Memsize-(MTU+8) */
    write_eth_reg16( ERXST, RX_START );
    write_eth_reg16( ERXND, RX_END );
    //write_eth_reg16( ERXWRPT, (RX_START | 0x01)+ 1);  /* this is automatically done by hardware but not reliably, RX_START must be 0 */
    write_eth_reg16( ERXRDPT, (RX_END - 1) | 0x01 );    /* we have read all packages up to ERXRDPT, so hardware is allowed to use the whole buffer, ERXRDPT must be uneven (errata rev. 4) */
    enc28j60_rx_packet_pointer = read_eth_reg16( ERXWRPT );
    
    /* MAC initialization */
    write_mac_reg( MACON2, 0x00 );
    write_mac_reg( MACON1, MARXEN | TXPAUS | RXPAUS );
    set_bits( MACON3, TXCRCEN | FRMLNEN ); // | PADCFG1 | PADCFG0
    /* configuer MTU */
    write_mac_reg16( MAMXFL, MTU );
    write_mac_reg16( MAIPG, 0x0c12 );
    /* write the mac address */
    write_mac_reg( MAADR0, this_ethernet_address.addr[0] );
    write_mac_reg( MAADR1, this_ethernet_address.addr[1] );
    write_mac_reg( MAADR2, this_ethernet_address.addr[2] );
    write_mac_reg( MAADR3, this_ethernet_address.addr[3] );
    write_mac_reg( MAADR4, this_ethernet_address.addr[4] );
    write_mac_reg( MAADR5, this_ethernet_address.addr[5] );
    /* set the recieve filter, accept UNICAST or BROADCAST. Only accept valid CRC packages */
//  write_eth_reg( ERXFCON, UCEN | BCEN);
    /* initialize the PHY */
    write_phy_reg16( PHCON2, HDLDIS );
    write_phy_reg16( PHLCON, 0x472 ); // LEDA_DISPLAY_LINK_STATUS | LEDB_DISPLAY_RECIEVE_ACTIVITY | STRETCH_139 | STRCH );
    /* duplex configuration */
    clear_bits( ECON1, TXRTS | RXEN );      //stop reception and transmission
    /* select full duplex operation */
    unsigned short pr = read_phy_reg16( PHCON1 );
    write_phy_reg16( PHCON1, pr | PDPXMD );
    pr = read_phy_reg16( PHCON1 );  /* set the MAC config according to the result */
    if ((pr & PDPXMD) != 0) {
        /* full duplex mode */
        write_mac_reg( MACON3, read_mac_reg( MACON3 ) | FULDPX );
        write_mac_reg( MABBIPG, 0x15 );
        DBG_I_S( "enc28j60_init(): Full Duplex operation.\n" );
    } else {
        /* half duplex mode */
        write_mac_reg( MACON3, read_mac_reg( MACON3 ) & (~FULDPX) );
        write_mac_reg( MABBIPG, 0x12 );
        DBG_I_S( "enc28j60_init(): Half Duplex operation.\n" );
    }
    if (link_status() == 0) {
        DBG_I_S( "enc28j60_init(): Link is down.\n" );
    }

    /* setup the external interrupt line inside the enc28j60 */
    write_eth_reg( EIE, INTIE | PKTIE );
    
    /* we could insert a delay here to wait for the enc to start up and set the connection (negotiation), but leaving
    this out results in some packet losts at the beginning, don't worry about that.
    */
    /* enable reception */
    set_bits( ECON1, RXEN );
    
    /* enable interrupt (external line) */
    ENC28j60_IRQ_SETUP();
    ENC28j60_IRQ_ENABLE();
    
    BENCHMARK_INIT();
            
    // return to normal mode:
    ENC28j60_PROTECT_UNCHECK();
    SPI_DESYNCHRONIZE( sync );
    
    return ERROR_SUCCESS;
}

ERROR enc28j60_power_down( void ) {
    // deinitialisation is taking place in synchronous spi mode
    // switch to synchronous mode:
    unsigned int sync = SPI_SYNCHRONIZE();
    // deinitialize the enc28j60. always use the spi_sync_* functions here!
    
    
    // return to normal mode:
    SPI_DESYNCHRONIZE( sync );
}

void enc28j60_irq_handler( void ) {
    // enable spi interrupt and disable external interrupt
    ENC28j60_IRQ_DISABLE();
    SPI_ENABLE_INTERRUPT();
}

void spi_interrupt_handler( void) {
    BENCHMARK_START();
    ENC28j60_PROTECT_CHECK();
    DBG_V_S( "{" );
    /* check what todo here. */
    /* statemachine for each packet. Any fifo can be read or written here. */
    /* use the spi_sync* functions here since the interrupt will only be called if it is synchronized! */
    static unsigned int state = 0;
    static unsigned int tlen, rlen;
    static void * tmem, * rmem;
    static unsigned char status_vector[6];
next_state:
    if (state == 0) {
        // IDLE.
        // check if there is a packet to be transmitted and start that if so,
        // if no packet to be transmitted watch out for recieved packets
        if ((enc28j60_txfifo.len > 0) && ((read_eth_reg( ECON1 ) & TXRTS) == 0)) {
            DBG_V_S ( "txs" );
            // there is a new packet to be transmitted and the transmitter is currently idle.
            // transmit the next packet
            tmem = enc28j60_txfifo.data[ enc28j60_txfifo.rpointer ].mem;
            tlen = enc28j60_txfifo.data[ enc28j60_txfifo.rpointer ].len;
            if (++enc28j60_txfifo.rpointer >= ENC28j60_FIFO_LEN) enc28j60_txfifo.rpointer = 0;
            enc28j60_txfifo.len--;
            // write the pointer registers for the ethernet side
            write_eth_reg16( ETXST, TX_START );
            write_eth_reg16( ETXND, TX_START + tlen - 4 /* subtract ethernet crc len */ + 1 - 1);
            // write the pointer registers for the spi side
            write_eth_reg16( EWRPT, TX_START );
            // write the buffer memory instruction
            spi_sync_start( enc28j60_dev );
            spi_sync_tmit( INS_WBM );
            // write the per-packet control byte (use default settings from MACON3)
            spi_sync_tmit( 0x0F );
            // transmit the total buffer, now asynchronously!
            spi_async_xfer( tmem, 0, tlen - 4 );
            state = 1;      // next state is packet transmitted!
            DBG_V_S( "t" ); DBG_V_I( tmem );
        } else if (read_eth_reg( EPKTCNT ) != 0){
            DBG_V_S ( "rxs" );
            // yes, new packets in buffer
            // set up the spi pointer registers in enc28j60
            write_eth_reg16( ERDPT, enc28j60_rx_packet_pointer );
            // read the status vector asynchronously...
            spi_sync_start( enc28j60_dev );
            spi_sync_tmit( INS_RBM );
            spi_sync_rmit();            // read back the value returned from the instruction
            spi_async_xfer( status_vector, status_vector, 6 );
            state = 2;              // next state is examining the status vector and starting the packet read.
        } else {
            if ((read_eth_reg( ECON1 ) & RXEN) == 0) {
                DBG_C_S( "enc28j60_interrupt_handler: packet reception disabled! reenabling again.\n" );
                // also check for the pointer register. this could be damaged, too
                enc28j60_rx_packet_pointer = read_eth_reg16( ERXWRPT ); // this discards any packages that are in the cache!
                unsigned short rxrdpt = (enc28j60_rx_packet_pointer - 1) | 1;
                if (enc28j60_rx_packet_pointer == 0)
                /* if the next packet will start at address 0x00, we must set rxrdpt to
                *  RX_END
                */
                    rxrdpt = (RX_END - 1) | 1;
                write_eth_reg16( ERXRDPT, rxrdpt );
                set_bits( ECON1, RXEN );
            }
            // nothing to do, neither reception nor transmission. disable spi interrupt
            SPI_DISABLE_INTERRUPT();
            state = 0;      // next state is idle again, interrupt is enabled if packets should be transmitted or int-line is active
            // enable the external line interrup (level sensitive!)
            ENC28j60_IRQ_ENABLE();
        }
    } else if (state == 1) {
        DBG_V_S ( "txe" );
        // packet has been transmitted, free the buffer associated with it and start the transmission inside
        // the enc28j60 (pointer in the enc already set up befor transfer of the packet itself.)
        // end the spi transfer
        spi_sync_end();
        mem_free( tmem );
        // start transmission in enc28j60
        set_bits( ECON1, TXRTS );
        state = 0;  // next state is idle
        goto next_state;   // handle next state immediatly, no asynchronous transfer active
    } else if (state == 2) {
        DBG_V_S ( "rx1" );
        // the status vector has been read, examine it:
        rlen = status_vector[2] | (status_vector[3]<<8);  // packet length
        if ((rlen > MTU) | ((status_vector[4] & (1<<7)) == 0)) {
            DBG_E_S( "enc28j60_interrupt_handler: invalid packet recieved.\n" );
            // if packet too long and/or crc and symbol errors, discard it
            spi_sync_end();
            goto packet_discard;
        }
        // get the needed memory
        rmem = mem_allocate( rlen );
        if (rmem != 0) {
            // spi transfer is still active, just go on reading the whole packet
            spi_async_xfer( rmem, rmem, rlen );
        } else {
            // no memory for allocation, discard the packet
            DBG_E_S( "enc28j60_interrupt_handler: no memory for packet.\n" );
            goto packet_discard;
        }
        state = 3;  // next state is packet recieved
        DBG_V_S( "r" ); DBG_V_I( rmem ); DBG_V_S( "-" ); DBG_V_I( rlen );
    } else if (state == 3) {
        DBG_V_S ( "rxe" );
        // packet has been read into rmem,rlen. now we can give the buffer to the rxfifo and
        // end transfer
        spi_sync_end();
        // clean up the registers in enc28j60
        if (enc28j60_rxfifo.len >= ENC28j60_FIFO_LEN) {
            // if the fifo is full, discard the packet
            DBG_E_S( "enc28j60_interrupt_handler: no free queue place.\n" );
            // free the buffer
            mem_free( rmem );
        } else {
            // add the packet to the rxfifo
            unsigned int i = enc28j60_rxfifo.rpointer + enc28j60_rxfifo.len;
            if (i >= ENC28j60_FIFO_LEN) i -= ENC28j60_FIFO_LEN;
            enc28j60_rxfifo.len++;
            enc28j60_rxfifo.data[i].mem = rmem;
            enc28j60_rxfifo.data[i].len = rlen;
        }
packet_discard:;
        // cleanup registers
        enc28j60_rx_packet_pointer = status_vector[0] | (status_vector[1]<<8);
        /* tell the controller that we have read up to here. */
        /* rxrdpt must be odd due to the bug in silicon revision 4 */
        unsigned short rxrdpt = (enc28j60_rx_packet_pointer - 1) | 1;
        if (enc28j60_rx_packet_pointer == 0)
        /* if the next packet will start at address 0x00, we must set rxrdpt to
         *  RX_END
         */
            rxrdpt = (RX_END - 1) | 1;
        write_eth_reg16( ERXRDPT, rxrdpt );
        /* decrement packet counter */
        set_bits( ECON2, PKTDEC );
        state = 0;  // next state s idle!
        goto next_state;    // no asynchronoustransfer active -> handle next state immediatly.
    } else {
        // this should not happen. state is not correct if this happens!
        DBG_E_S( "!" );
        state = 0;
        goto next_state;    // no asynchronoustransfer active -> handle next state immediatly.
    }
    BENCHMARK_STOP();
    ENC28j60_PROTECT_UNCHECK();
    DBG_V_S( "}" );
}

