/*

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Copyright (C) 2006 by Clemens Helfmeier,
    Clemens(at)Helfmeier[dot]de

*/


/* enc28j60.c
 *
 * this file contains functions to access register and memory on the enc28j60.
 * it relays on the spi-library.
 *
 * by:   Clemens Helfmeier
 * date: 13.06.2006
 */



#include "net/mac/enc28j60_register.h"
#include "spi.h"
#include "system.h"

#include "config_spi.h"
#define enc28j60_dev	SPI_1

#include "debug/debug.h"
#define DBG_LEVEL   DBG_LEVEL_ERROR
#include "debug/debug.h"

static unsigned char current_bank = 0xFF;

void set_bank( unsigned char bank) {
	/* bank doesn't need to be set for register higher than 0x1A*/
    if ((bank & 0x1F) > 0x1A) return;
	bank = bank>>5 & 0x03;
	if (current_bank == bank) return;
    DBG_V_S( "set_bank(): " ); DBG_V_B( bank ); DBG_V_S( ".\n" );
    spi_sync_start( enc28j60_dev );
	spi_sync_tmit( INS_BFC | (ECON1 & 0x1F) );
	spi_sync_tmit( ((~bank) & 0x03 ) );
	spi_sync_end();
	spi_sync_start( enc28j60_dev );
	spi_sync_tmit( INS_BFS | (ECON1 & 0x1F) );
	spi_sync_tmit( ((bank) & 0x03 ) );
	spi_sync_end();
	current_bank = bank;
}

/* ETHERNET Register modifying functions */
unsigned char read_eth_reg( unsigned char addr ) {
    DBG_V_S( "read_eth_reg(): " ); DBG_V_B( addr ); DBG_V_S( " ... -> " );
	set_bank( addr );
    spi_sync_start( enc28j60_dev );
	spi_sync_tmit( INS_RCR | (addr & 0x1F) );
    spi_sync_tmit( 0xFF );
    spi_sync_rmit();
	unsigned char retval = spi_sync_rmit();
	spi_sync_end();
    DBG_V_B( retval ); DBG_V_S( ".\n" );
	return retval;
}
void write_eth_reg( unsigned char addr, unsigned char data ) {
    DBG_V_S( "write_eth_reg(): " );  DBG_V_B( addr ); DBG_V_S( " " ); DBG_V_B( data ); DBG_V_S( ".\n" );
	set_bank( addr );
	spi_sync_start( enc28j60_dev );
	spi_sync_tmit( INS_WCR | (addr & 0x1F) );
	spi_sync_tmit( data );
	spi_sync_end();
}

/* MII Register modifying functions */
unsigned char read_mii_reg( unsigned char addr ) {
    DBG_V_S( "read_mii_reg(): " ); DBG_V_B( addr ); DBG_V_S( ".\n" );
	set_bank( addr );
    spi_sync_start( enc28j60_dev );
	spi_sync_tmit( INS_RCR | (addr & 0x1F) );
	spi_sync_tmit( 0xFF );	// write dummy byte 
    spi_sync_tmit( 0xFF );  // write reading byte 
	unsigned char retval = spi_sync_rmit();	// read data byte 
	spi_sync_end();
	return retval;
}

/* MAC Register modifying functions */

/* PHY Register modifying functions */
unsigned short read_phy_reg16( unsigned char addr ) {
	/* read the corresponding PHY register */
    DBG_V_S( "read_phy_reg16(): " ); DBG_V_B( addr ); DBG_V_S( ".\n" );
	write_mii_reg( MIREGADR, addr & 0x1F );
	write_mii_reg( MICMD, MIIRD );
	/* wait for read to finish */
	delay_us(10);
	while ((read_mii_reg( MISTAT ) & BUSY) != 0);
	/* clear the command */
	write_mii_reg( MICMD, 0x00 );
	return read_mii_reg16( MIRD );
}
void write_phy_reg16( unsigned char addr, unsigned short data ) {
	/* write the correct phy register */
    DBG_V_S( "write_phy_reg16(): " ); DBG_V_B( addr ); DBG_V_S( " " ); DBG_V_I( data ); DBG_V_S( ".\n" );
    write_mii_reg( MIREGADR, addr & 0x1F );
	write_mii_reg16( MIWR, data );
	/* wait for phy idle */
	delay_us(10);
	while ((read_mii_reg( MISTAT ) & BUSY) != 0);
}

/* MEMORY modifying functiosn */
void write_mem_sync( void *data, unsigned short addr, unsigned short len ) {
    static unsigned short last_addr = -1;
	/* write the address */
    DBG_V_S( "write_mem(): " ); DBG_V_I( data ); DBG_V_S( " " ); DBG_V_I( addr ); DBG_V_S( " " ); DBG_V_I( len ); DBG_V_S( ".\n" );
    if (last_addr != addr) {
	   write_eth_reg16( EWRPT, addr );
    }
    last_addr = addr + len;
	/* write the data */
    spi_sync_start( enc28j60_dev );
	spi_sync_tmit( INS_WBM );
	for (unsigned int i = 0; i<len; i++)
		spi_sync_tmit( *((unsigned char*)data++) );
	spi_sync_end();
}

void read_mem_sync( void *data, unsigned short addr, unsigned short len ) {
    static unsigned short last_addr = -1;
	/* write the address */
    DBG_V_S( "read_mem(): " ); DBG_V_I( data ); DBG_V_S( " " ); DBG_V_I( addr ); DBG_V_S( " " ); DBG_V_I( len ); DBG_V_S( ".\n" );
    if (last_addr != addr) {
	   write_eth_reg16( ERDPT, addr );
    }
    last_addr = addr + len;
	/* read the data */
    spi_sync_start( enc28j60_dev );
	spi_sync_tmit( INS_RBM );
    spi_sync_tmit( 0xFF );      // transmit reading byte
    spi_sync_rmit();            // read the answer to ins-byte
    for (unsigned int i = len-1; i>0; i--) {
        if (i > 1) spi_sync_tmit( 0xFF );
        *((unsigned char*)data++) = spi_sync_rmit();    // read all answers
    }
	/* and finaly deselect the chip */
	spi_sync_end();
}


/* BIT modifying functions */
void clear_bits( unsigned char addr, unsigned char bits ) {
	/* clear the bits that are set in variable bits */
    DBG_V_S( "clear_bits(): " ); DBG_V_B( addr ); DBG_V_S( " " ); DBG_V_B( bits ); DBG_V_S( ".\n" );
    set_bank( addr );
	spi_sync_start( enc28j60_dev );
	spi_sync_tmit( INS_BFC | (addr & 0x1F) );
	spi_sync_tmit( bits );
	spi_sync_end();
}
void set_bits( unsigned char addr, unsigned char bits ) {
	/* set the bits that are set in variable bits */
    DBG_V_S( "set_bits(): " ); DBG_V_B( addr ); DBG_V_S( " " ); DBG_V_B( bits ); DBG_V_S( ".\n" );
	set_bank( addr );
    spi_sync_start( enc28j60_dev );
	spi_sync_tmit( INS_BFS | (addr & 0x1F) );
	spi_sync_tmit( bits );
	spi_sync_end();
}

/* MISC functions */
void soft_reset( void ) {
    DBG_V_S( "soft_reset():\n" );
    spi_sync_start( enc28j60_dev );
    spi_sync_tmit( INS_SWRST );
    spi_sync_end();
}

unsigned int link_status( void ) {
    DBG_V_S( "link_status():\n" );
	/* return 0 if link is down. */
	if ((read_eth_reg( ESTAT ) & CLKRDY) == 0) return 0;	// link is down if clock is off
	if ((read_eth_reg( ECON2 ) & (PWRSV | VRPS)) != 0) return 0;	// link is down if VoltageReg or Logic in power down.
	/* now it is possible to access phy register */
	if ((read_phy_reg16( PHSTAT1 ) & LLSTAT) == 0) return 0;
	return 1;	/* link is up */
}
