/*

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Copyright (C) 2006 by Clemens Helfmeier,
    Clemens(at)Helfmeier[dot]de

*/


#include <string.h>

#include "net/netconfig.h"
#include "net/hostconfig.h"
#include "net/netutils.h"
#include "net/proto/icmp.h"

#include "debug/debug.h"
#define DBG_LEVEL   DBG_LEVEL_ERROR
#include "debug/debug.h"
        
/** include reassembling module */
#include "ipv4_reass.c"
        

/** initialize the ip layer */
ERROR ipv4_init( void ) {
    ERROR ret = 0;
    ret = ipv4_reass_init();
    if (ret != ERROR_SUCCESS) return ret;
    return ERROR_SUCCESS;
}

/** send a packet to dst on protocol proto */
ERROR ipv4_send_packet( void * mem, unsigned int len, ipv4_address dst, ipv4_protocol proto ) {
    // the counter is incremented with each sent packet.
    static uint16_t ipv4_tx_id_counter = 0;
    DBG_V_S("ipv4_send_packet: " ); DBG_V_I( len ); DBG_V_S( ".\n" );
    ipv4_header * ihdr = (ipv4_header *)(mem + IPV4_OFFSET - IPV4_HEADER_SIZE);
    ihdr->ihl = IPV4_HEADER_SIZE / 4;
    ihdr->version = 4;
    ihdr->tos = 0;
    ihdr->tot_len = htons(len - (IPV4_TOTAL_SIZE - IPV4_HEADER_SIZE - IPV4_TAIL_SIZE));  //size of ip packet = total size - header - tail
    ihdr->id = htons(ipv4_tx_id_counter++);
    ihdr->frag_off = HTONS(0);
    ihdr->ttl = 128;
    ihdr->protocol = proto;
    ihdr->check = 0;
    ihdr->saddr = htonl(this_ipv4_address);
    ihdr->daddr = htonl(dst);
    // calculate checksum
    ihdr->check = htons( IPV4_CRC( ihdr, IPV4_HEADER_SIZE ) );
    // resolve the ipv4 address to the ethernet address
    ethernet_address edest;
    ERROR ret = IPV4_RESOLVE( &edest, dst );
    if (ret == ERROR_SUCCESS) {
        // send packet
        return IPV4_SEND_PACKET( mem, len, &edest );
    } else {
        // return an error.
        return ret;
    }
}

/** handle an incoming packet */
void ipv4_input_handler( void * mem, unsigned int len ) {
    ipv4_header * ihdr = (ipv4_header *)(mem + IPV4_OFFSET - IPV4_HEADER_SIZE);
    // make some plausabillity checks
    if ((ihdr->version != 4) || (ihdr->ttl == 0)) {
        // discard packet
        DBG_V_S( "ipv4_input_handler: packet with invalid version received.\n" );
        return;
    }
    // is the packet for us?
    ipv4_address dst;
    dst = ntohl(ihdr->daddr);
    if (dst != this_ipv4_address) {
        // discard packet
        DBG_V_S( "ipv4_input_handler: packet received which is not for us.\n" );
        return;
    }
    // if there are options, just remove them
    if (ihdr->ihl != IPV4_HEADER_SIZE/4) {
        DBG_V_S( "ipv4_input_handler: skipping options.\n" );
        // remove the options by moving the packet forward
        unsigned int dl = ihdr->ihl*4 - IPV4_HEADER_SIZE;
        memmove( mem + IPV4_OFFSET, mem + IPV4_OFFSET + dl, len - IPV4_OFFSET - dl );
        len -= dl;
    }
    // if the memory is longer than the actual ipv4 packet, reduce len (possibly added padding by ETHERNET)
    unsigned int dl = len - (ntohs(ihdr->tot_len) + IPV4_TOTAL_SIZE - IPV4_HEADER_SIZE - IPV4_TAIL_SIZE);
    if (dl > 0) {
        DBG_V_S( "ipv4_input_handler: removed " ); DBG_V_I( dl ); DBG_V_S( " bytes of padding from packet.\n" );
        len -= dl;
    } else if (dl < 0) {
        DBG_E_S( "ipv4_input_handler: packet's length field is invalid by "); DBG_E_I( -dl ); DBG_E_S( " bytes.\n" );
    }
    /* test for fragmented packet */
    if ((ntohs(ihdr->frag_off) & (FRAGMENT_OFFSET | FRAGMENT_MOREFRAGMENTS)) != 0) {
        // handle fragments separatly.
        ipv4_reass_input_handler( mem, len );
    } else {
        // this is is not fragmented 
        if (ihdr->protocol == IPV4_PID_UDP) {
            UDP_INPUT_HANDLER( mem, len );
//        } else if (ihdr->protocol == IPV4_PID_TCP) {
//            TCP_INPUT_HANDLER( mem, len );
        } else if (ihdr->protocol == IPV4_PID_ICMP) {
            ICMP_INPUT_HANDLER( mem, len );
        } else {
            // no handler given...
            DBG_V_S( "ipv4_input_handler: no handler for packet.\n" );
            icmp_dstunr( mem, len, ICMP_DSTUNR_PROTO );
        }
    }
}
