/*

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Copyright (C) 2006 by Clemens Helfmeier,
    Clemens(at)Helfmeier[dot]de

*/


/** configuration file for the network stack.
 * This network stack is linked statically, it uses static sizes of
 * packages and thus has very high efficience.
 * 
 * A packet always consists of zero or more header and a variable length
 * payload and zero or more tails.
 * The size of a packet is always predictable - not at runtime but at design time!
 * Thus the size of a packet must be calculatable by a macro.
 * For each protocol there is the header and tail size defined as macro and a
 * total size which includes all underlying layer's header and tail sizes.
 * Additional there has to be a definition for the actual beginning of the layers
 * payload data in the packet (ipv4_offset ==  beginning of payload of ipv4 packet, e.g. udp header).
 * 
 * @author: Clemens Helfmeier
 * @date:   29.10.2006
 */

#ifndef _NETCONFIG_H_
#define _NETCONFIG_H_


/** NETWORK STACK LAYOUT:
 * MAC HEADER (ENC28j60 HEADER if needed.)
 * + ARP HEADER/PACKET
 * + ETHERNET HEADER + TAIL
 * | + IPV4 HEADER
 * |   + UDP HEADER
 * |   + TCP HEADER
 * |   + ICMP HEADER
 */

// --- MAC LAYER ---
#define MAC_HEADER_SIZE         (0)
#define MAC_TAIL_SIZE           (0)
// no underlying layer
#define MAC_TOTAL_SIZE          (0)
#define MAC_OFFSET              (MAC_HEADER_SIZE)

// --- ETHERNET LAYER ---
#define ETHERNET_HEADER_SIZE    (14)
#define ETHERNET_TAIL_SIZE      (4)
#define ETHERNET_TOTAL_SIZE     (MAC_TOTAL_SIZE + ETHERNET_HEADER_SIZE + ETHERNET_TAIL_SIZE)
#define ETHERNET_OFFSET         (MAC_OFFSET + ETHERNET_HEADER_SIZE)

// --- IPV4 LAYER ---
// the ipv4 header size is variable, we don't implement options so it
// is always fixed length of 20 bytes
#define IPV4_HEADER_SIZE        (20)
#define IPV4_TAIL_SIZE          (0)
#define IPV4_TOTAL_SIZE         (ETHERNET_TOTAL_SIZE + IPV4_HEADER_SIZE + IPV4_TAIL_SIZE)
#define IPV4_OFFSET             (ETHERNET_OFFSET + IPV4_HEADER_SIZE)

// --- UDP LAYER ---
#define UDP_HEADER_SIZE         (8)
#define UDP_TAIL_SIZE           (0)
#define UDP_TOTAL_SIZE          (IPV4_TOTAL_SIZE + UDP_HEADER_SIZE + UDP_TAIL_SIZE)
#define UDP_OFFSET              (IPV4_OFFSET + UDP_HEADER_SIZE)

// --- ICMP LAYER ---
#define ICMP_HEADER_SIZE        (8)
#define ICMP_TAIL_SIZE          (0)
#define ICMP_TOTAL_SIZE         (IPV4_TOTAL_SIZE + ICMP_HEADER_SIZE + ICMP_TAIL_SIZE)
#define ICMP_OFFSET             (IPV4_OFFSET + ICMP_HEADER_SIZE)


// --- ARP LAYER ---
// ARP has 8 static bytes and a variable number of bytes in the packet - depending on
// underlying layers (ethernet + ipv4 = 20 Byte)
#define ARP_HEADER_SIZE         (8)
#define ARP_TAIL_SIZE           (0)
#define ARP_TOTAL_SIZE          (ETHERNET_TOTAL_SIZE + ARP_HEADER_SIZE + ARP_TAIL_SIZE)
#define ARP_OFFSET              (ETHERNET_OFFSET + ARP_HEADER_SIZE)

// include common host<->network functions
#include "net/hostconfig.h"

// includes of all layers:
#include "net/dll/ethernet.h"
#include "net/mac/enc28j60.h"
#include "net/net/ipv4.h"
#include "net/proto/arp.h"
#include "net/proto/udp.h"
#include "net/proto/tcp.h"
#include "net/proto/icmp.h"

// include definitions of addresses and layer specific configurations...
#include "net/userconfig.h"

// connecting the layers...
/** all layers are linked statically. The linkage is defined here.
 * output of packets is quite obvious: For packet oriented protocols the
 * LAYER_SEND_PACKET( data, len, dst,  ) macro calls the underlying layer to send the packet.
 * Depending on the procotol, a layer may also require a LAYER_RESOLVE macro which translates
 * the address of LAYER to the address of underlying layers.
 * Packet reception is a bit more difficult. Here we need to adjust the decision procedure
 * of the underlying layer. 
 */

// ---- PROTOCOL DEFINTIONS ----
// MAC: no protocol field in MAC header
// ETHERNET: protocol field is either IPV4 or ARP
#define ETHERNET_PID_IPV4       HTONS(0x0800)
#define ETHERNET_PID_ARP        HTONS(0x0806)
// IPV4: protocol field ist either ICMP, UDP or TCP
#define IPV4_PID_UDP            (17)
#define IPV4_PID_TCP            (6)
#define IPV4_PID_ICMP           (1)

// ---- PACKET OUTPUT ----
// ETHERNET: packets are given to the mac (enc28j60 here...) The CRC in the tail is automatically calculated (padded).
#define ETHERNET_SEND_PACKET( data, len )       enc28j60_send_packet( data, len )
// ARP: Packets are send via ethernet layer, always broadcast packets
#define ARP_SEND_PACKET( data, len, dst )       ethernet_send_packet( data, len, dst, ETHERNET_PID_ARP )
// IPV4: Packets are given to the ethernet layer, dst is the ethernet address here!
#define IPV4_SEND_PACKET( data, len, dst )      ethernet_send_packet( data, len, dst, ETHERNET_PID_IPV4 )
// UDP: Packets are given to the ipv4 layer, dst is a ipv4-address here.
#define UDP_SEND_PACKET( data, len, dst )       ipv4_send_packet( data, len, dst, IPV4_PID_UDP )
// ICMP: Packets are given to the ipv4 layer, dst is a ipv4-address here.
#define ICMP_SEND_PACKET( data, len, dst )      ipv4_send_packet( data, len, dst, IPV4_PID_ICMP )
// TCP: Packets are given to the ipv4 layer, dst is a ipv4-address here.
#define TCP_SEND_PACKET( data, len, dst )       ipv4_send_packet( data, len, dst, IPV4_PID_TCP )

// ---- PACKET INPUT ----
// begin to read from the last entry! (MAC) the compiler complains about not defined macros in inline functions,
// so the layers are reverse ordered here (like packets are handled, bottom up)
// ICMP: packets are handled inside the icmp module
#define ICMP_INPUT_HANDLER( data, len )         //icmp_input_handler( data, len )
// TCP: packets are handled internally in the socket api...
#define TCP_INPUT_HANDLER( data, len )          //tcp_input_handler( data, len )
// UDP: packets are handled internally in the socket api...
#define UDP_INPUT_HANDLER( data, len )          //udp_input_handler( data, len )
// IPV4: packets received by the ipv4 layer are handeld by it and given on later. (reassembling, etc)
#define IPV4_INPUT_HANDLER( data, len )         //ipv4_input_handler( data, len )
// ARP: packets recieved by arp are always handled by the ARP itself (last layer in chain)
#define ARP_INPUT_HANDLER( data, len )          arp_input_handler( data, len )
// ETHERNET: packets recieved via ethernet are handled by either IPV4 or ARP, this is done by an inline function
#define ETHERNET_INPUT_HANDLER( data, len )     ethernet_input_handler( data, len )
static inline void ethernet_input_handler( void * data, unsigned int len ) {
    ethernet_header * ehdr = (ethernet_header*)(data + ETHERNET_OFFSET - ETHERNET_HEADER_SIZE);
    if (ehdr->proto == ETHERNET_PID_IPV4) {
        IPV4_INPUT_HANDLER( data, len );
    } else if (ehdr->proto == ETHERNET_PID_ARP) {
        ARP_INPUT_HANDLER( data, len );
    }
}
// MAC: packets recieved by the mac are always handled by the ethernet layer
#define MAC_INPUT_HANDLER( data, len )          ETHERNET_INPUT_HANDLER( data, len )

// ---- ADDRESS TRANSLATION ----
// MAC: no address translation needed
// ETHERNET: no address translation needed
// ARP: no translation needed
// IPV4: translation from ipv4 address to ethernet address needed
#define IPV4_RESOLVE( ha, pa )                  arp_resolve( ha, pa )
// UDP: no translation needed
// TCP: no translation needed
// ICMP: no translation needed

// ---- CRC CALCULATION ----
// IPV4: crc is usually calculated over the finished header... normal crc is accurate
#define IPV4_CRC( mem, len )                    mem_crc16( mem, len )
// ICMP: crc is calculated over the header and payload
#define ICMP_CRC( mem, len )                    mem_crc16( mem, len )
// UDP: crc is calculated over the payload and a pseudo-ip-header. the
// mem,len specifies the 
#define UDP_CRC( mem, len, dst, src )   ({DBG_E_S( "UDP_CRC not defined.\n" ); 0; })


// ---- ADDRESSES OF PACKETS ----
// MAC: has no src address (that is in ethernet address, dll)
//ETHERNET: Address is in the header field
#define ETHERNET_SRC( mem )                     (((ethernet_header*)(mem + ETHERNET_OFFSET - ETHERNET_HEADER_SIZE))->src)
// IPV4: Address is in the header field
#define IPV4_SRC( mem )                         (((ipv4_header*)(mem + IPV4_OFFSET - IPV4_HEADER_SIZE))->saddr)
// ARP: Src address is in the arp packet or ethernet packet...
#define ARP_SRC( mem )                          (((arp_header*)(mem + ARP_OFFSET - ARP_HEADER_SIZE))->sha)
// UDP: ontop of ipv4 -> ipv4-address
#define UDP_SRC( mem )                          IPV4_SRC( mem )
// TCP: ontop of ipv4 -> ipv4-address
#define TCP_SRC( mem )                          IPV4_SRC( mem )
// ICMP: ontop of ipv4 -> ipv4-address
#define ICMP_SRC( mem )                         IPV4_SRC( mem )






#endif

