/*

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Copyright (C) 2006 by Clemens Helfmeier,
    Clemens(at)Helfmeier[dot]de

*/


#include <stdint.h>
#include "net/netutils.h"
#include "net/netconfig.h"

#include "debug/debug.h"
#define DBG_LEVEL   DBG_LEVEL_VERBOSE
#include "debug/debug.h"

/** CRC CHECKSUM TAKEN FROM LWIP (both routines): */
/*
 * Copyright (c) 2001-2004 Swedish Institute of Computer Science.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
 * SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * This file is part of the lwIP TCP/IP stack.
 *
 * Author: Adam Dunkels <adam@sics.se>
 *
 */
/**
 * lwip checksum
 *
 * @param dataptr points to start of data to be summed at any boundary
 * @param len length of data to be summed
 * @return host order (!) lwip checksum (non-inverted Internet sum) 
 *
 * @note accumulator size limits summable length to 64k
 * @note host endianess is irrelevant (p3 RFC1071)
 */
uint16_t mem_crc16( void * dataptr, unsigned int len ) {
    // calculate the internet checksum of 16 bits
  uint32_t acc;
  uint16_t src;
  uint8_t *octetptr;

  acc = 0;
  /* dataptr may be at odd or even addresses */
  octetptr = (uint8_t*)dataptr;
  while (len > 1)
  {
    /* declare first octet as most significant
       thus assume network order, ignoring host order */
    src = (*octetptr) << 8;
    octetptr++;
    /* declare second octet as least significant */
    src |= (*octetptr);
    octetptr++;
    acc += src;
    len -= 2;
  }
  if (len > 0)
  {
    /* accumulate remaining octet */
    src = (*octetptr) << 8;
    acc += src;
  }
  /* add deferred carry bits */
  acc = (acc >> 16) + (acc & 0x0000ffffUL);
  if ((acc & 0xffff0000) != 0) {
    acc = (acc >> 16) + (acc & 0x0000ffffUL);
  }
  /* This maybe a little confusing: reorder sum using htons()
     instead of ntohs() since it has a little less call overhead.
     The caller must invert bits for Internet sum ! */
  return ~(uint16_t)acc;
}

uint16_t mem_crc16_pseudo(void * mem, unsigned int len, ipv4_address dst, ipv4_address src, ipv4_protocol proto, uint16_t proto_len)
{
  uint32_t acc;

    DBG_V_S( "mem_crc16_pseudo: " ); DBG_V_I( src ); DBG_V_S( " -> " ); DBG_V_I( dst ); DBG_V_S( " proto " ); DBG_V_I( proto ); DBG_V_S( " protolen " ); DBG_V_I( proto_len ); DBG_V_S( ".\n" );

    acc = (~mem_crc16(mem, len)) & 0x0FFFF;
  
    DBG_V_S( "mem_crc16_pseudo: memory " ); DBG_V_I( acc ); DBG_V_S( " result " );

    acc += ((src) & 0x0ffff);
    acc += (((src) >> 16) & 0x0ffff);
    acc += ((dst) & 0x0ffff);
    acc += (((dst) >> 16) & 0x0ffff);
    acc += (uint32_t)((uint16_t)proto);
    acc += (uint32_t)(proto_len);

    while ((acc & 0xFFFF0000) != 0) {
        acc = (acc & 0x0FFFF) + (acc >> 16);
    }
    
    DBG_V_I( ~acc & 0x0FFFF ); DBG_V_S( ".\n" );
    return (uint16_t)(~acc & 0x0FFFF);
}































