/*

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Copyright (C) 2006 by Clemens Helfmeier,
    Clemens(at)Helfmeier[dot]de

*/


#include <string.h>
#include <stdint.h>


#include "system/time.h"
#include "system/error.h"
#include "net/netconfig.h"
#include "net/proto/arp.h"

#include "memory/mem.h"

#include "debug/debug.h"
#define DBG_LEVEL   DBG_LEVEL_ERROR
#include "debug/debug.h"

#define hardware_address        ethernet_address
#define network_address         ipv4_address
#define this_hardware_address   this_ethernet_address
#define this_network_address    this_ipv4_address
#define this_network_mask       this_ipv4_mask
#define this_network_subnet     this_ipv4_subnet
#define this_network_gateway    this_ipv4_gateway
#define NETWORK_ADDRESS_TYPE    HTONS(0x0800)
#define HARDWARE_ADDRESS_TYPE   HTONS(0x0001)



/** Every recieved ARP-Entry is saved in the array. Old entries are overwritten
 * if the array is full.
 */
typedef struct {
    hardware_address ha;
    network_address pa;
    int time;       ///< An empty entry has time 0, all other entries have their recieve time saved here (positive) or their request time (negative).
} arp_entry;

/** ARP Packet definition
 */
typedef struct __attribute__ ((__packed__)) {
    uint16_t htype;
    uint16_t ptype;
    uint8_t hlen;
    uint8_t plen;
    uint16_t oper;
    hardware_address sha;
    network_address spa;
    hardware_address tha;
    network_address tpa;
} arp_packet;


#define ARP_CACHE_SIZE      2
arp_entry arp_cache[ ARP_CACHE_SIZE ];

/** initialize the ARP module
 * this also registers the input handler.
 */
ERROR arp_init( void ) {
    // clear the cache
    for (unsigned int i = 0; i< ARP_CACHE_SIZE; i++) {
        arp_cache[i].time = 0;
    }
    return ERROR_SUCCESS;
}

/**
 * Input function for arp packets. 
 */
void arp_input_handler( void * mem, unsigned int len ) {
    // copy packet to packet structure
    arp_packet * ap = (arp_packet *)(mem + ARP_OFFSET - ARP_HEADER_SIZE);
    // is this packet for the layer we expect it to be?
    if (((ap->ptype) == NETWORK_ADDRESS_TYPE) && (ap->plen == sizeof(network_address))) {
        // is it a query or an answer?
        if (ap->oper == HTONS(0x02)) {
            // answer
            network_address na = ntohl(ap->spa);
            // have a look if this entry is is in the cache (has been queried by us)
            for (unsigned int i = 0; i<ARP_CACHE_SIZE; i++) {
                if (arp_cache[i].pa == na) {
                    // this is the entry, save everything
                    arp_cache[i].ha = ap->sha;
                    arp_cache[i].time = time_getTickCount();
                    DBG_V_S( "arp_input_handler: answer recieved from " ); DBG_V_P( na ); DBG_V_S( ".\n" );
                    break;
                }
            }
        } else if (ap->oper == HTONS( 0x01 )) {
            // query
            DBG_V_S( "arp_input_handler: query recieved.\n" );
            // answer to the query if possible.
            network_address na = ntohl(ap->tpa);
            if (na == this_network_address) {
                // we are asked for our hardware address
                unsigned int buflen = ARP_TOTAL_SIZE - ARP_HEADER_SIZE + sizeof(arp_packet);
                char * buf = mem_allocate(buflen);
                arp_packet * rap = (arp_packet *)(buf + ARP_OFFSET - ARP_HEADER_SIZE);
                // fill the packet
                rap->tha = ap->sha;
                rap->tpa = ap->spa;
                rap->sha = this_hardware_address;
                rap->spa = htonl(this_network_address);
                rap->oper = HTONS( 0x02 );      // answer.
                rap->htype = (HARDWARE_ADDRESS_TYPE);
                rap->ptype = (NETWORK_ADDRESS_TYPE);
                rap->hlen = sizeof( hardware_address );
                rap->plen = sizeof( network_address );
                DBG_I_S( "arp_query: sending answer packet ... " );
                ARP_SEND_PACKET( buf, buflen, &rap->tha );
                DBG_I_S( "done.\n" );
                mem_free( buf );
            }
        } // if oper == 0x01
    } // if packet is arp vor ipv4
}

void arp_query( network_address pa ) {
    // query for the given address
    unsigned int buflen = ARP_TOTAL_SIZE - ARP_HEADER_SIZE + sizeof(arp_packet);
    char * buf = mem_allocate(buflen);
    arp_packet * ap = (arp_packet *)(buf + ARP_OFFSET - ARP_HEADER_SIZE);

    // fill the packet
    ap->tha = ethernet_broadcast_address;
    ap->tpa = htonl(pa);
    ap->sha = this_hardware_address;
    ap->spa = htonl(this_network_address);
    ap->oper = HTONS( 0x01 );
    ap->htype = (HARDWARE_ADDRESS_TYPE);
    ap->ptype = (NETWORK_ADDRESS_TYPE);
    ap->hlen = sizeof( hardware_address );
    ap->plen = sizeof( network_address );
    
    DBG_I_S( "arp_query: sending packet ... " );
    ARP_SEND_PACKET( buf, buflen, &ap->tha );
    DBG_I_S( "done.\n" );
    
    mem_free( buf );
    
    arp_entry * ae = 0;
    for (unsigned int i = 0; i<ARP_CACHE_SIZE; i++) {
        if (arp_cache[i].pa == pa) {
            // we have sent for this entry, mark the time
            ae = &arp_cache[i];
            break;
        }
        if ((((arp_cache[i].time == 0) || (arp_cache[i].time > 20000000)) && (ae == 0)) || ((ae != 0) && (ae->time < arp_cache[i].time))) {
            // this entry is either the first we find and very old, or it is older than anyone find before.
            // it can not be found if we already have the origintal entry.
            ae = &arp_cache[i];
        }
    }
    if (ae == 0) {
        // cache can be full
        DBG_E_S( "arp_query: cache damaged when querying for " ); DBG_E_I( pa ); DBG_E_S( ".\n" );
        arp_init(); // reinitialize cache
        return;
    }
    ae->pa = pa;
    ae->time = -time_getTickCount();
    if (ae->time == 0) ae->time = 1;
    DBG_V_S( "arp_query: time = " ); DBG_V_I( ae->time ); DBG_V_S( ".\n" );
}

ERROR arp_resolve( hardware_address * ha, network_address pa ) {
    // have a look in the cache in case we already have the address
    // is address rewriting needed? this would be the case if the target ip is not in our network
    if ((pa & this_network_mask) != this_network_subnet) {
        // send the packet to the hardware addess of this_network_gateway instead of pa.
        pa = this_network_gateway;
    }
    for (unsigned int i = 0; i<ARP_CACHE_SIZE; i++) {
        if (arp_cache[i].pa == pa) {
            if (arp_cache[i].time > 0) {
                DBG_V_S( "arp_resolve: entry in cache: " ); DBG_V_I( pa ); DBG_V_S( ".\n" );
                *ha = arp_cache[i].ha;
                return ERROR_SUCCESS;
            } else if (arp_cache[i].time < 0) {
                // the query was already placed
                DBG_V_S( "arp_resolve: entry not yet answered: " ); DBG_V_I( pa ); DBG_V_S( " time: " ); DBG_V_I( arp_cache[i].time); DBG_V_S( " current time: " ); DBG_V_I( time_getTickCount() ); DBG_V_S( ".\n" );
                
                if (arp_cache[i].time + (int)time_getTickCount() < 2000000)
                    // this has been less than 2000ms before now. don't send again.
                    return ERROR_ARP_NOT_IN_CACHE;
                else {
                    DBG_V_S( "apr_resolve: time has passed: " ); DBG_V_I( arp_cache[i].time + (int)time_getTickCount() ); DBG_V_S( ".\n" );
                }
            }
            // else time == 0
            break;
        }
    }
    // not in cache, send query.
    DBG_V_S( "arp_resolve: entry not in cache: " ); DBG_V_I( pa ); DBG_V_S( ".\n" );
    arp_query( pa );
    // return failure
    return ERROR_ARP_NOT_QUERYING;
}



