/*

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Copyright (C) 2006 by Clemens Helfmeier,
    Clemens(at)Helfmeier[dot]de

*/


#include "system/error.h"
#include "net/netconfig.h"
#include "net/proto/tcp.h"

#include "memory/mem.h"
#include "system/time.h"

#include <stdint.h>
#include <string.h>

#include "debug/debug.h"
#define DBG_LEVEL   DBG_LEVEL_VERBOSE
#include "debug/debug.h"


typedef struct __attribute__ ((__packed__)){
    tcp_port srcport;
    tcp_port dstport;
    uint32_t seqnr;
    uint32_t acknr;
    uint16_t flags;
    uint16_t window;
    uint16_t checksum;
    uint16_t urgent_ptr;
    // the tcp header can be followed by any number of 4 byte
    // aligned options.
} tcp_header;

// flags field:
#define TCP_FLAG_FIN            HTONS((1<<0))
#define TCP_FLAG_SYN            HTONS((1<<1))
#define TCP_FLAG_RST            HTONS((1<<2))
#define TCP_FLAG_PSH            HTONS((1<<3))
#define TCP_FLAG_ACK            HTONS((1<<4))
#define TCP_FLAG_URG            HTONS((1<<5))
#define TCP_FLAG_DATAOFS(flags) (((flags) >> 2) & 0x3C)
// we do not send any options, our header is always 4*5=20byte in length
#define TCP_FLAG_DOFS           HTONS((0x5000)) 

// any socket in the tcp stack must be remembered for packet reception:
#define TCP_SOCKET_CACHE_SIZE   4
tcp_socket * tcp_socket_cache[TCP_SOCKET_CACHE_SIZE];

// flags for the tcp_socket.flags field
#define TCP_SOCKET_SENDACK      (1<<0)
#define TCP_SOCKET_SENDSYN      (1<<1)

#define MSL                     2*60*1000000
// connection timeout 5min.
#define RX_TIMEOUT              5*60*1000000
// ack send timeout (delayed ack) 1sec.
#define ACK_TIMEOUT             1*1000000

ERROR tcp_init( void ) {
    memset( tcp_socket_cache, 0x00, sizeof( tcp_socket_cache ) );
    return ERROR_SUCCESS;
}

/** send a packet with payload and flags, seqnr, acknr.
 */
ERROR tcp_send_packet( tcp_socket * s, void * mem, unsigned int len, uint16_t tcpflags, uint32_t seqnr, uint32_t acknr ) {
    // send a standard packet. fill in the tcp header and give it to the next layer.
    // the memory will not be freed!
    if (mem == 0) {
        return ERROR_FAILURE;
    }
    tcp_header * hdr = (tcp_header*)(mem + TCP_OFFSET - TCP_HEADER_SIZE);
    hdr->srcport = s->localport;
    hdr->dstport = s->remoteport;
    hdr->seqnr = htonl(seqnr);
    hdr->acknr = htonl(acknr);
    hdr->flags = ((tcpflags & (TCP_FLAG_SYN | TCP_FLAG_RST | TCP_FLAG_ACK | TCP_FLAG_URG | TCP_FLAG_FIN | TCP_FLAG_PSH)) | TCP_FLAG_DOFS);
    hdr->checksum = 0;
    hdr->window = htons(s->recv.window);
    hdr->urgent_ptr = 0;
    // calculate checksum
    hdr->checksum = htons(TCP_CRC( hdr, len - TCP_TOTAL_SIZE + TCP_HEADER_SIZE, s->remote, this_ipv4_address, len - TCP_TOTAL_SIZE + TCP_HEADER_SIZE));
    DBG_V_S( "tcp_send_packet: " ); DBG_V_I( s->remote ); DBG_V_S( ":" ); DBG_V_I( s->remoteport );
    DBG_V_S( " len " ); DBG_V_I( len ); DBG_V_S( " " );
    if ((hdr->flags & TCP_FLAG_SYN) != 0) DBG_V_S( "SYN," );
    if ((hdr->flags & TCP_FLAG_ACK) != 0) DBG_V_S( "ACK," );
    if ((hdr->flags & TCP_FLAG_RST) != 0) DBG_V_S( "RST," );
    if ((hdr->flags & TCP_FLAG_FIN) != 0) DBG_V_S( "FIN," );
    DBG_V_S( " seqnr "); DBG_V_I( ntohl(hdr->seqnr) );
    DBG_V_S( " acknr "); DBG_V_I( ntohl(hdr->acknr) );
    DBG_V_S( " window "); DBG_V_I( ntohl(hdr->window) );
    DBG_V_S( ".\n" );
    ERROR ret = TCP_SEND_PACKET( mem, TCP_TOTAL_SIZE, s->remote );
    if (ret == ERROR_SUCCESS) {
        if ((hdr->flags & TCP_FLAG_ACK) != 0) {
            // ack sent, clear flag
            s->flags &= ~TCP_FLAG_ACK;
            s->ack_timeout = time_getTickCount() + ACK_TIMEOUT;
        }
    }
    return ret;
}

/** send a packet without payload, only containing flags and sequence/ack numbers
 */
ERROR tcp_send_flag_packet( tcp_socket * s, uint16_t tcpflags, uint32_t seqnr, uint32_t acknr ) {
    // generate a new tcp message containing the RST flag and
    // seqnr as sequence number
    void * mem = mem_allocate( TCP_TOTAL_SIZE );    // no payload.
    if (mem == 0) {
        // memory allocation failed, return
        return ERROR_FAILURE;
    }
    ERROR ret = tcp_send_packet( s, mem, TCP_TOTAL_SIZE, tcpflags, seqnr, acknr );
    mem_free( mem );
    return ret;
}


ERROR tcp_remove_acked_segments( tcp_socket * s) {
    DBG_V_S( "tcp_remove_acked_semgents: NOT IMPLEMENTED.\n" );
}

ERROR tcp_reorder_input( tcp_socket * s, void * mem, unsigned int len ) {
    // packets that are received and contain segment text, go into this routine.
    // find an empty place in recv.queue
    for (unsigned int i = 0; i<RECV_QUEUE_LEN; i++) {
        if (s->recv.queue[i].mem == 0) {
            s->recv.queue[i].mem = mem;
            s->recv.queue[i].len = len;
            // the memory is in use now.
            mem_use( mem );
            DBG_V_S( "tcp_reorder_input: saving packet to index " ); DBG_V_I( i ); DBG_V_S( ": " ); DBG_V_I( ((tcp_header*)(s->recv.queue[i].mem+TCP_OFFSET-TCP_HEADER_SIZE))->seqnr ); DBG_V_S( ", " ); DBG_V_I( s->recv.nextseq ); DBG_V_S( ".\n" );
            return ERROR_SUCCESS;
        }
    }
    return ERROR_TCP_NO_CACHE;
}

// ffind the next packet that should be handled and put the ack on the wire
ERROR tcp_reorder_output( tcp_socket * s, void ** mem, unsigned int * len ) {
    for (unsigned int i = 0; i<RECV_QUEUE_LEN; i++) {
        if ((s->recv.queue[i].mem != 0) && (((tcp_header*)(s->recv.queue[i].mem+TCP_OFFSET-TCP_HEADER_SIZE))->seqnr == s->recv.nextseq)) {
            // this is the next packet to be handled:
            *mem = s->recv.queue[i].mem;
            *len = s->recv.queue[i].len;
            // send the ack
            s->recv.nextseq = ((tcp_header*)(s->recv.queue[i].mem+TCP_OFFSET-TCP_HEADER_SIZE))->seqnr + s->recv.queue[i].len - TCP_TOTAL_SIZE;
            s->flags |= TCP_FLAG_ACK;
            // remove the entry from the queue
            s->recv.queue[i].mem = 0;
            DBG_V_S( "tcp_reorder_output: packet read, " ); DBG_V_I( s->recv.nextseq ); DBG_V_S( ".\n" );
            return ERROR_SUCCESS;
        } else if (s->recv.queue[i].mem != 0) {
            DBG_V_S( "tcp_reorder_output: packet did not match sequence number: " ); DBG_V_I( ((tcp_header*)(s->recv.queue[i].mem+TCP_OFFSET-TCP_HEADER_SIZE))->seqnr ); DBG_V_S( " != " ); DBG_V_I( s->recv.nextseq ); DBG_V_S( ".\n" );
        }
    }
    return ERROR_TCP_NO_CACHE;
}

void tcp_input_handler( void * mem, unsigned int len ) {
    tcp_header * hdr = (tcp_header*)(mem + TCP_OFFSET - TCP_HEADER_SIZE);
    if (len < TCP_TOTAL_SIZE) {
        // packet is too short, the full tcp header does not fit.
        DBG_E_S( "tcp_input_handler: packet received which is too short.\n" );
        return;
    }
    tcp_address src = TCP_SRC( mem );

    DBG_V_S( "tcp_input_handler: len " ); DBG_V_I( len - TCP_TOTAL_SIZE + TCP_HEADER_SIZE ); DBG_V_S( " flags " ); DBG_V_I( hdr->flags ); DBG_V_S( " src " ); DBG_V_I( src ); DBG_V_S( " remoteport " ); DBG_V_I( hdr->srcport ); DBG_V_S( " localport " ); DBG_V_I( hdr->dstport ); DBG_V_S( " seqnr "); DBG_V_I( ntohl(hdr->seqnr) ); DBG_V_S( " acknr "); DBG_V_I( ntohl(hdr->acknr) ); DBG_V_S( ".\n" );
    // check if crc is valid:
    uint16_t crc = TCP_CRC( hdr, len - TCP_TOTAL_SIZE + TCP_HEADER_SIZE, this_ipv4_address, src, len - TCP_TOTAL_SIZE + TCP_HEADER_SIZE);
    if (crc != 0) {
        DBG_E_S( "tcp_input_handler: packet received which has incorrect TCP checksum: " ); DBG_E_I( crc ); DBG_E_S( ".\n" );
        return;
    }
    tcp_socket * s = 0;
    // have look for the corresponding socket
    for (unsigned int i = 0; i<TCP_SOCKET_CACHE_SIZE; i++) {
        if (tcp_socket_cache[i] != 0) {
            tcp_socket * s2 = tcp_socket_cache[i];
            // check for correct port for this socket
            if (s2->localport != hdr->dstport) {
                DBG_V_S( "tcp_input_handler: socket port does not match.\n" );
                continue;
            }
            // check for correct ip addrss
            if (s2->remote != (s2->msk & src)) {
                DBG_V_S( "tcp_input_handler: src addresses do not match.\n" );
                continue;
            }
            // check for correct remote port if state is not listening
            if ((s2->state != TCP_LISTEN) && (s2->remoteport != hdr->srcport)) {
                DBG_V_S( "tcp_input_handler: remote ports do not match.\n" );
                continue;
            }
            // this socket corresponds to the packet.
            DBG_V_S( "tcp_input_handler: socket found.\n" );
            s = s2;
            break;
        }
    }
    if ((s == 0)) {
        // no socket found for this port, discard the packet
        DBG_V_S( "tcp_input_handler: no socket for packet from " ); DBG_V_I( src ); DBG_V_S( ":" ); DBG_V_I( hdr->srcport ); DBG_V_S( " to localhost:" ); DBG_V_I( hdr->dstport ); DBG_V_S( ".\n" );
        goto drop_segment;
    }
    DBG_V_S( "tcp_input_handler: handling packet... \n" );
    // convert seqnr and acknr and calculate length of payload (segment)
    hdr->acknr = ntohl( hdr->acknr );
    hdr->seqnr = ntohl( hdr->seqnr );
    // remove any optional options
    if (TCP_FLAG_DATAOFS(hdr->flags) > TCP_HEADER_SIZE) {
        unsigned int dl = (TCP_FLAG_DATAOFS(hdr->flags) - TCP_HEADER_SIZE);
        DBG_V_S( "tcp_input_handler: removing " ); DBG_V_I( dl ); DBG_V_S( " bytes of options.\n" );
        memmove( ((char*)hdr) + TCP_HEADER_SIZE, ((char *)(hdr)) + TCP_HEADER_SIZE + dl, len - TCP_TOTAL_SIZE - dl);
        len -= dl;
        hdr->flags = (hdr->flags & ~HTONS(0xF000)) | TCP_FLAG_DOFS;
    }
    // payload size
    unsigned int seglen = len - TCP_TOTAL_SIZE - TCP_HEADER_SIZE;
    // payload pointer
    char * payload = hdr+1;
    
    // the following "script" is taken from RFC793 "SEGMENT ARRIVES"
    if (s->state == TCP_CLOSED) {
        if ((hdr->flags & TCP_FLAG_RST) == 0) {
            if ((hdr->flags & TCP_FLAG_ACK) != 0) {
                tcp_send_flag_packet( s, TCP_FLAG_RST, hdr->acknr, 0 );
            } else {
                tcp_send_flag_packet( s, TCP_FLAG_RST | TCP_FLAG_ACK, 0, hdr->seqnr + seglen );
            }
            DBG_V_S( "tcp_input_handler: reset send due to CLOSED state.\n" );
        } else {
            // do not answer a RST with a RST, this would lead to a dead lock.
            DBG_V_S( "tcp_input_handler: socket CLOSED.\n" );
        }
        goto drop_segment;
    } else if (s->state == TCP_LISTEN) {
        // 1) check for RST, if present, discard the packet
        if ((hdr->flags & TCP_FLAG_RST) != 0) goto drop_segment;
        // 2) check for ACK
        if ((hdr->flags & TCP_FLAG_ACK) != 0) {
            // any ack in listen state is bad, respond with seq=hdr->acknr,RST
            // the ip address of the socket must be overwritten now...
            ipv4_address tmp = s->remote;
            tcp_port tmp2 = s->remoteport;
            s->remote = src;
            s->remoteport = hdr->srcport;
            tcp_send_flag_packet( s, TCP_FLAG_RST, hdr->acknr, 0 );
            s->remote = tmp;
            s->remoteport = tmp2;
            DBG_V_S( "tcp_input_handler: ACK received in listen state. RST sent.\n" );
            goto drop_segment;
        }
        // 3) check for SYN
        if ((hdr->flags & TCP_FLAG_SYN) != 0) {
            // check security. if security does not match, send a reset.
            // here: security always matches.
            DBG_V_S( "tcp_input_handler: accepting connection.\n" );
            s->recv.iseq = hdr->seqnr;  // this was seqnr, but it seems that all tcp programs use acknr here!
            s->recv.nextseq = s->recv.iseq + 1;   // initial sequence number received.
            // fill in any unspeciefied data of the socket:
            s->remote = src;
            s->remoteport = hdr->srcport;
            // the socket is now bound to that one ip:
            s->msk = A2IPV4( 255,255,255,255 );
            // calculate our iss and acknowledge by sending an apropriate packet
            s->send.iseq = 100 /* iss */ + 1;
            s->send.nextseq = s->send.iseq + 1;
            s->send.unackseq = s->send.iseq;
            tcp_send_flag_packet( s, TCP_FLAG_SYN | TCP_FLAG_ACK, s->send.iseq, s->recv.nextseq );
            // new connection state is SYN_RECEIVED:
            s->state = TCP_SYN_RECEIVED;
            // if there is anything attached to thispacket, we should keep it.
            // it is to be processed in SYN_RECEIVED state.
            if ((seglen != 0) || ((hdr->flags & (TCP_FLAG_FIN | TCP_FLAG_URG | TCP_FLAG_PSH)) != 0)) {
                // keep the segment for later processing
                goto queue_segment;
            }
            // ok, segment processed
            goto processed_segment;
        }
        // any left segment is dropped.
        goto drop_segment;
    } else if (s->state == TCP_SYN_SENT) {
        // 1) check for ACK
        if ((hdr->flags & TCP_FLAG_ACK) != 0) {
            if (((s->send.iseq - hdr->seqnr >= 0) || (hdr->seqnr - s->send.nextseq > 0)) && ((hdr->flags & TCP_FLAG_RST) == 0)) {
                // send a reset
                tcp_send_flag_packet( s, TCP_FLAG_RST, hdr->acknr, 0 );
                goto processed_segment;
            }
            if (!((hdr->seqnr - s->send.unackseq >= 0) && (s->send.nextseq - hdr->seqnr >= 0))) {
                // ack is not acceptable
                goto drop_segment;
            }
        }
        // here: ack was acceptable or not set
        // 2) check for RST bit
        if ((hdr->flags & TCP_FLAG_RST) != 0) {
            // enter closed state and drop the segment
            s->state = TCP_CLOSED;
            goto drop_segment;
        }
        // 3) security check and precedence
        /*
        if (security check fails) {
            if ((hdr->flags & TCP_FLAG_ACK) != 0) {
                tcp_send_flag_packet( s, TCP_FLAG_RST, hdr->acknr, 0 );
            } else {
                tcp_send_flag_packet( s, TCP_FLAG_RST | TCP_FLAG_ACK, 0, hdr->seqnr + seglen );
            }
            goto drop_segment;
        }
        */
        // 4) check for SYN
        if ((hdr->flags & TCP_FLAG_SYN) != 0) {
            s->recv.iseq = hdr->seqnr;
            s->recv.nextseq = s->recv.iseq + 1;
            if ((hdr->flags & TCP_FLAG_ACK) != 0) {
                DBG_V_S( "tcp_input_handler: advancing send.unackseq.\n" );
                // advance the acknowledged sequence numbers.
                s->send.unackseq = hdr->acknr;
                // and handle the packets in our sending queue (remove all ack'ed ones.)
                tcp_remove_acked_segments( s );
            }
            if (s->send.unackseq - s->send.iseq > 0) {
                DBG_V_S( "tcp_input_handler: SYN has been ACKed.\n" );
                // our SYN has been acked.
                s->state = TCP_ESTABLISHED;
                tcp_send_flag_packet( s, TCP_FLAG_ACK, s->send.nextseq, s->recv.nextseq );
            } else {
                DBG_V_S( "tcp_input_handler: sent SYN, ACK. new state is SYN_RECEIVED." );
                // new state is SYN_RECEIVED
                s->state = TCP_SYN_RECEIVED;
                tcp_send_flag_packet( s, TCP_FLAG_ACK | TCP_FLAG_SYN,  s->send.iseq, s->recv.nextseq );
            }
            // queue any left information of this packet
            goto queue_segment;
        }
        // 5) neither syn nor rst is set > drop the packet
        goto drop_segment;
    } else {
        // for any other state of the socket, go on like this:
        // 1) check sequence number for all states
        if (s->recv.window == 0) {
            if (seglen > 0) goto drop_segment;
            else if (hdr->seqnr != s->recv.nextseq) goto drop_segment;
        } else {
            if (!((hdr->seqnr - s->recv.nextseq >= 0) && (s->recv.nextseq + s->recv.window - hdr->seqnr > 0))) {
                if ((seglen == 0) || (!((hdr->seqnr + seglen - 1 - s->recv.nextseq >= 0) && (s->recv.nextseq + s->recv.window - (hdr->seqnr + seglen - 1) > 0)))) {
                    // seq nr not acceptable, send ack if rst is not set
                    if ((hdr->flags & TCP_FLAG_RST) == 0)
                        tcp_send_flag_packet( s, TCP_FLAG_ACK, s->send.nextseq, s->recv.nextseq );
                    // then drop the unacceptable segment
                    DBG_V_S( "tcp_input_handler: unacceptable sequence number " ); DBG_V_I( hdr->seqnr ); DBG_V_S( ".\n" );
                    goto drop_segment;
                }
            }
        }
        // sequence number is acceptable
        // 2) check for RST
        if ((hdr->flags & TCP_FLAG_RST) != 0) {
            // the connection was aborted or refused.
            DBG_E_S( "tcp_input_handler: connection refused or aborted. application NOT informed.\n" );
            s->state = TCP_CLOSED;
            // remove any segments
            tcp_remove_acked_segments( s );
            goto drop_segment;
        }
        // 3) check for security and precedence
        // security always matches.
        // 4) check for SYN
        if ((hdr->flags & TCP_FLAG_SYN) != 0) {
            // reset the connection
            tcp_send_flag_packet( s, TCP_FLAG_RST, s->send.nextseq, s->recv.nextseq );
            // new state isclosed
            s->state = TCP_CLOSED;
            DBG_V_S( "tcp_input_handler: SYN received, connection CLOSED.\n" );
            // drop the segment
            goto drop_segment;
        }
        // 5) check for ACK
        if ((hdr->flags & TCP_FLAG_ACK) != 0) {
            if ((hdr->acknr - s->send.unackseq >= 0) && (s->send.nextseq - hdr->acknr >= 0)) {
                // acknr is acceptable
                if (s->state == TCP_SYN_RECEIVED) {
                    s->state = TCP_ESTABLISHED;
                }
                // for any other state, the handling is quite the same:
                // update unackseqnr
                s->send.unackseq = hdr->acknr;
                // remove acked segments now...
                tcp_remove_acked_segments( s );
                // go to next state:
                if (s->state == TCP_FIN_WAIT1) s->state = TCP_FIN_WAIT2;
                else if (s->state == TCP_FIN_WAIT2) {
                    // tell the user that close is successful...
                    DBG_E_S( "tcp_input_handler: NOT IMPLEMENTED: user close is successful.\n" );
                    // new state is closed
                    s->state = TCP_CLOSED;
                } else if (s->state == TCP_CLOSING) {
                    // check if this ack ackes our FIN
                    DBG_E_S( "tcp_input_handler: NOT IMPLEMENTED: only go to TIME_WAIT state if ack is for our FIN.\n" );
                    // if so, go to TIME_WAIT
                    s->state_timeout = time_getTickCount() + MSL;
                    s->state = TCP_TIME_WAIT;
                } else if (s->state == TCP_LAST_ACK) {
                    // ack of our FIN received... ( TODO: maybe also check this!)
                    DBG_E_S( "tcp_input_handler: ACK for FIN received, socket CLOSED.\n" );
                    s->state = TCP_CLOSED;
                } else if (s->state == TCP_TIME_WAIT) {
                    // resend the ACK and restart timeout
                    tcp_send_flag_packet( s, TCP_FLAG_FIN | TCP_FLAG_ACK, s->send.nextseq, s->recv.nextseq );
                    s->state_timeout = time_getTickCount() + MSL;
                    DBG_E_S( "tcp_input_handler: TIME_WAIT: ack sent.\n" );
                }
                if (s->state != TCP_ESTABLISHED) {
                    // if state is not established, we processed the sequence.
                    goto processed_segment;
                }
            } else if (s->state == TCP_SYN_RECEIVED) {
                // acknr is not acceptable and we are in TCP_SYN_RECEIVED state, send a reset
                tcp_send_flag_packet( s, TCP_FLAG_RST, hdr->acknr, 0 );
                goto processed_segment;
            } else if (s->send.unackseq - hdr->acknr > 0) {
                // duplicate ack, ignore
                goto drop_segment;
            } else if (hdr->acknr - s->send.nextseq > 0) {
                // ack for someting not yet sent...
                // the full syntax for this ack is not described in RFC793
                tcp_send_flag_packet( s, TCP_FLAG_ACK, s->send.nextseq, s->recv.nextseq );
            }
        }
        // 6) check for URG
        if ((hdr->flags & TCP_FLAG_URG) != 0) {
            // TODO: implement handling of the urgent data pointer...
        }
        // 7) process the segment text
        // update our window pointer
        s->send.window = hdr->window;
        if (((s->state == TCP_ESTABLISHED) || (s->state == TCP_FIN_WAIT1) || (s->state == TCP_FIN_WAIT2)) && (seglen > 0)) {
            tcp_reorder_input( s, mem, len );
        } else {
            // ignore the text from this segment
            if (seglen != 0) DBG_E_S( "tcp_input_handler: segment text ignored (" ); DBG_V_I( seglen ); DBG_V_S( " bytes).\n" );
        }
        // 8) check for FIN
        if ((hdr->flags & TCP_FLAG_FIN) != 0) {
            if ((s->state == TCP_CLOSED) || (s->state == TCP_LISTEN) || (s->state == TCP_SYN_SENT)) {
                // drop the fin if in wrong state
                DBG_V_S( "tcp_input_handler: FIN received in wrong state.\n" );
                goto drop_segment;
            }
            s->recv.nextseq = hdr->seqnr + seglen + 1;  // advance over the FIN
            // send ack for the fin
            tcp_send_flag_packet( s, TCP_FLAG_ACK | TCP_FLAG_FIN, s->send.nextseq, s->recv.nextseq );
            DBG_V_S( "tcp_input_handler: FIN received, ack sent.\n" );
            // select next state
            if (s->state == TCP_FIN_WAIT1) {
                if ((hdr->flags & TCP_FLAG_ACK) != 0) {
                    // our FIN has been ACKed, enter TIME_WAIT
                    s->state = TCP_TIME_WAIT;
                    s->state_timeout = time_getTickCount() + MSL;
                    DBG_V_S( "tcp_input_handler: new state is TIME_WAIT.\n" );
                } else {
                    s->state = TCP_CLOSING;
                    DBG_V_S( "tcp_input_handler: new state is CLOSING.\n" );
                }
            } else if ((s->state == TCP_SYN_RECEIVED) || (s->state == TCP_ESTABLISHED)) {
                s->state = TCP_CLOSE_WAIT;
                DBG_V_S( "tcp_input_handler: new state is CLOSE_WAIT.\n" );
            } else if (s->state == TCP_FIN_WAIT2) {
                s->state = TCP_TIME_WAIT;
                s->state_timeout = time_getTickCount() + 2*MSL;
                DBG_V_S( "tcp_input_handler: new state is TIME_WAIT.\n" );
            }
        }
        // 9) packet handled.
        goto processed_segment;
    }
    
queue_segment:
        DBG_E_S( "tcp_input_handler: NOT IMPLEMENTED: keeping packets for later.\n" );
        return;
processed_segment:
        // any segments that have been processed will go to here
        // update connection timeout
        if (s->state != TCP_TIME_WAIT) {
            s->state_timeout = time_getTickCount() + RX_TIMEOUT;
        }
        DBG_V_S( "tcp_input_handler: segment processed.\n" );
        return;
drop_segment:
        // any dropped segments will end here.
        DBG_V_S( "tcp_input_handler: segment dropped.\n" );
        return;
}
    
/** listen for connection requests on port localport from any host that matches src/msk
 * if a connection request is recieved, the connection is accepted.
 */
ERROR tcp_accept( tcp_socket * s, tcp_address remote, tcp_address msk, tcp_port localport ) {
    memset( s, 0x00, sizeof(tcp_socket) );
    // initialize the socket
    s->state = TCP_LISTEN;
    s->remote = remote;
    s->msk = msk;
    s->localport = htons(localport);
    s->send.window = 20;
    s->recv.window = 20;
    unsigned int e = -1;
    // add the socket to the socket cache
    for (unsigned int i = 0; i<TCP_SOCKET_CACHE_SIZE; i++) {
        if (tcp_socket_cache[i] == s) {
            DBG_V_S( "tcp_accept: socket already in socket cache.\n" );
            return ERROR_SUCCESS;
        }
        if (tcp_socket_cache[i] == 0) {
            e = i;
        }
    }
    if (e != -1) {
        tcp_socket_cache[e] = s;
        return ERROR_SUCCESS;
    }
    // no free place in socket cache
    return ERROR_TCP_NO_CACHE;
}

/** terminate a tcp connection
 */
ERROR tcp_close( tcp_socket * s ) {
    // close the connection
    if ((s->state == TCP_CLOSE_WAIT) || (s->state == TCP_ESTABLISHED) || (s->state == TCP_SYN_RECEIVED)) {
        // send a FIN packet
        tcp_send_flag_packet( s, TCP_FLAG_FIN, s->send.nextseq, 0 );
        // go to the next state:
        if (s->state == TCP_CLOSE_WAIT) s->state = TCP_LAST_ACK;
        else s->state = TCP_FIN_WAIT1;
        return ERROR_SUCCESS;
    } else
        return ERROR_TCP_INVALID_STATE;
}

/** send a packet into the tcp connection
 * the connection must be established!
 */
ERROR tcp_send( tcp_socket * s, void * mem, unsigned int len );

/** recieve a packet from a tcp connection
 * connection must be established.
 */
ERROR tcp_recv( tcp_socket * s, void ** mem, unsigned int * len ) {
    return tcp_reorder_output( s, mem, len );
}

/** handle timed tcp events.
 */
void tcp_timer( void ) {
    // cycle through open sockets
    for (unsigned int i = 0; i<TCP_SOCKET_CACHE_SIZE; i++) {
        if ((tcp_socket_cache[i] != 0) && (tcp_socket_cache[i]->state != TCP_CLOSED)) {
            tcp_socket * s = tcp_socket_cache[i];
            uint32_t time = time_getTickCount();
            if ((s->state == TCP_CLOSED) || (s->state == TCP_LISTEN)) continue;
            if (time > s->state_timeout) {
                // we didn't receive something for a long time, close the connection
                DBG_V_S( "tcp_timer: closing socket...\n" );
                tcp_close( s );
                s->state = TCP_CLOSED;
                continue;
            }
            if ((time > s->ack_timeout) && ((s->flags & TCP_FLAG_ACK) != 0)) {
                // we didn't acknowledge quite long, do this now
                DBG_V_S( "tcp_timer: sending ACK.\n" );
                tcp_send_flag_packet( s, TCP_FLAG_ACK, s->send.nextseq, s->recv.nextseq );
            }
        }
    }
}



