/*

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Copyright (C) 2006 by Clemens Helfmeier,
    Clemens(at)Helfmeier[dot]de

*/





#ifndef _TCP_H_
#define _TCP_H_

#include <stdint.h>

#include "system/error.h"

#ifdef __cplusplus
extern "C" {
#endif

        
typedef uint16_t tcp_port;
typedef ipv4_address tcp_address;


typedef enum {
    TCP_CLOSED,
    TCP_LISTEN,
    TCP_SYN_SENT,
    TCP_SYN_RECEIVED,
    TCP_ESTABLISHED,
    TCP_FIN_WAIT1,
    TCP_FIN_WAIT2,
    TCP_CLOSE_WAIT,
    TCP_CLOSING,
    TCP_LAST_ACK,
    TCP_TIME_WAIT,
} tcp_connection_state;

// maximum number of unacknowledged packets that are sent
#define SEND_QUEUE_LEN      10
// maximum of packets that can be received out of order
#define RECV_QUEUE_LEN      10

typedef struct {
    tcp_connection_state state;
    // ports and addresses are all stored in network order here!
    // conversation takes place in the listen and open functions
    tcp_address remote;     ///< remote ip address
    tcp_address msk;        ///< netmask for listening sockets
    tcp_port localport;     ///< local port
    tcp_port remoteport;    ///< remote port
    
    struct {
        uint32_t iseq;      ///< initial sequence number
        uint32_t unackseq;  ///< first unacknowledged sequence number
        uint32_t nextseq;   ///< sequence number for next data to be sent
        uint32_t window;    ///< seqence number window size. unack...unack+wnd numbers are allowed to be sent
        
        // output queue. Any packet sent that carries information, must be put into this queue. The queue is automatically
        // emptied when ACKs are received.
        struct {
            void * mem;
            unsigned int len;
        } queue[ SEND_QUEUE_LEN ];
    } send;
    struct {
        uint32_t iseq;      ///< initial sequence number
        uint32_t nextseq;   ///< sequence numbers which are allowed for reception
        uint32_t window;    ///< reception window size
        
        // received packets are put into this queue if they are received. when the user call tcp_recv(..) the packets
        // are removed from the queue.
        struct {
            void * mem;
            unsigned int len;
        } queue[ RECV_QUEUE_LEN ];
    } recv;
    uint32_t flags;         ///< flags which are important for timer. e.g. sending delayed acks, etc
    uint32_t ack_timeout;         ///< the time_getTickCount() value when the next ack should be sent (timeout).
    uint32_t state_timeout;    ///< time when the 2 * MSL end (TIME_WAIT state), or when the next packet must be received (any other state)

    
} tcp_socket;


ERROR tcp_init( void );

void tcp_input_handler( void * mem, unsigned int len );


/** listen for connection requests on port localport from any host that matches src/msk
 * if a connection request is recieved, the connection is accepted and the src and msk are
 * adjusted, the socket is then assocciated with that one connection.
 */
ERROR tcp_accept( tcp_socket * s, tcp_address src, tcp_address msk, tcp_port localport );

/** terminate a tcp connection
 */
ERROR tcp_close( tcp_socket * s );

/** send a packet into the tcp connection
 * the connection must be established!
 */
ERROR tcp_send( tcp_socket * s, void * mem, unsigned int len );

/** recieve a packet from a tcp connection
 * connection must be established.
 */
ERROR tcp_recv( tcp_socket * s, void ** mem, unsigned int * len );

/** determine the connection state
 */
static inline tcp_connection_state tcp_state( tcp_socket * s ) {
    return s->state;
}

void tcp_timer( void );



#ifdef __cplusplus
}
#endif
#endif
