/*

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Copyright (C) 2006 by Clemens Helfmeier,
    Clemens(at)Helfmeier[dot]de

*/


#include <string.h>

#include "system/error.h"
#include "net/netconfig.h"
#include "net/proto/udp.h"
#include "net/proto/icmp.h"
#include "memory/mem.h"

        
#include "debug/debug.h"
#define DBG_LEVEL       DBG_LEVEL_VERBOSE
#include "debug/debug.h"


udp_socket * sockets[ UDP_SOCKET_COUNT ];

ERROR udp_init( void ) {
    memset( sockets, 0x00, sizeof( sockets ) );
    return ERROR_SUCCESS;
}

void udp_input_handler( void * mem, unsigned int len ) {
    udp_header * uhdr = (udp_header*)(mem + UDP_OFFSET - UDP_HEADER_SIZE);
    // is the crc valid?
#if UDP_DOCHECKSUM
    if ((uhdr->crc != 0x00) && (UDP_CRC(uhdr, len - UDP_TOTAL_SIZE + UDP_HEADER_SIZE) != 0)) {
        // invalid crc, discard packet
        DBG_I_S( "udp_input_handler: invalid CRC packet received.\n" );
        return;
    }
#endif
    if (ntohs(uhdr->len) + UDP_TOTAL_SIZE - UDP_HEADER_SIZE - UDP_TAIL_SIZE != len) {
        DBG_I_S( "udp_input_handler: invalid length packet received.\n" );
        return;
    }
    // find the corresponding socket:
    for (unsigned int i = 0; i<UDP_SOCKET_COUNT; i++) {
        udp_address psrc;
        psrc = UDP_SRC( mem );
        if ((sockets[i]->remote == (sockets[i]->msk & psrc)) /* packet is from the correct remote side */
             && (sockets[i]->localport == uhdr->dst_port) /* ports are the same */
             && (sockets[i]->remoteport == uhdr->src_port)) {
            if (sockets[i]->qlen >= UDP_QUEUE_LEN) {
                DBG_I_S( "udp_input_handler: no space in queue. discarding packet.\n" );
                return;
            }
            // calculate windex
            unsigned int windex = sockets[i]->rindex + sockets[i]->qlen;
            if (windex >= UDP_QUEUE_LEN) windex -= UDP_QUEUE_LEN;
            // add the packet
            sockets[i]->buffer[windex].mem = mem;
            sockets[i]->buffer[windex].len = len;
            sockets[i]->qlen++;
            // the packet is in use now (keep it):
            mem_use( mem );
            // finished
            return;
        }
    }
    DBG_V_S( "udp_input_handler: no handler for packet.\n" );
    icmp_dstunr( mem, len, ICMP_DSTUNR_PORT );
}


/** open a new socket for listening */
ERROR udp_listen( udp_socket * s, udp_address remote, udp_address msk, udp_port localport, udp_port remoteport ) {
    // set this socket into listening state
    // is the socket already in the buffer?
    unsigned int fs = -1;    ///< remember a free socket entry
    for (unsigned int i = 0; i<UDP_SOCKET_COUNT; i++) {
        if (sockets[i] == s) {
            fs = i;
            break;
        }
        if ((sockets[i] == 0) && (fs == -1)) {
            fs = i;
        }
    }
    if (fs == -1) {
        // socket not in queue and no free socket there
        DBG_E_S( "udp_listen: out of sockets.\n" );
        return ERROR_UDP_NO_CACHE;
    }
    // fs is the socket
    sockets[fs] = s;
    // initialize s
    s->remote = remote & msk;
    s->msk = msk;
    s->localport = htons(localport);
    s->remoteport = htons(remoteport);
    s->rindex = 0;
    s->qlen = 0;
    memset( s->buffer, 0x00, sizeof( s->buffer ) );
    // now we are listening on s
    return ERROR_SUCCESS;
}

/** read from a socket
 * on success, the application must free the returned buffer
 */
ERROR udp_recv( udp_socket * s, void ** mem, unsigned int * len ) {
    // rindex and windex are always called from the main routine, no synchronisation needed
    if (s->qlen != 0) {
        // return this buffer entry and advance the buffer pointer
        *mem = s->buffer[s->rindex].mem;
        *len = s->buffer[s->rindex].len;
        if (++s->rindex >= UDP_QUEUE_LEN) s->rindex = 0;
        s->qlen--;
        return ERROR_SUCCESS;
    } else {
        // no packet in queue
        return ERROR_UDP_EMPTY;
    }
}

/** write to a socket */
ERROR udp_send( udp_socket * s, void * mem, unsigned int len ) {
    if (len < UDP_TOTAL_SIZE) return ERROR_FAILURE;
    udp_header * uhdr = (udp_header*)(mem + UDP_OFFSET - UDP_HEADER_SIZE);

    uhdr->dst_port = s->remoteport;
    uhdr->src_port = s->localport;
    uhdr->len = htons(len - UDP_TOTAL_SIZE + UDP_HEADER_SIZE + UDP_TAIL_SIZE);
    uhdr->crc = 0;
#if UDP_DOCHECKSUM
    uhdr->crc = htons( UDP_CRC( mem, len, s->remote, this_udp_address ) );
#endif
    return UDP_SEND_PACKET( mem, len, s->remote );
}


